"""Variant definitions for Query 7."""

from __future__ import annotations

from benchbox.core.tpchavoc.variant_base import StaticSQLVariant

VARIANTS = {
    1: StaticSQLVariant(
        1,
        "Scalar subqueries: Use scalar subqueries for volume calculation",
        """\

select
    supp_nation,
    cust_nation,
    l_year,
    (select sum(l2.l_extendedprice * (1 - l2.l_discount))
     from supplier s2, lineitem l2, orders o2, customer c2, nation n3, nation n4
     where s2.s_suppkey = l2.l_suppkey
       and o2.o_orderkey = l2.l_orderkey
       and c2.c_custkey = o2.o_custkey
       and s2.s_nationkey = n3.n_nationkey
       and c2.c_nationkey = n4.n_nationkey
       and n3.n_name = shipping.supp_nation
       and n4.n_name = shipping.cust_nation
       and extract(year from l2.l_shipdate) = shipping.l_year
       and l2.l_shipdate between date '1995-01-01' and date '1996-12-31') as revenue
from (
    select distinct
        n1.n_name as supp_nation,
        n2.n_name as cust_nation,
        extract(year from l_shipdate) as l_year
    from
        supplier, lineitem, orders, customer, nation n1, nation n2
    where
        s_suppkey = l_suppkey
        and o_orderkey = l_orderkey
        and c_custkey = o_custkey
        and s_nationkey = n1.n_nationkey
        and c_nationkey = n2.n_nationkey
        and ((n1.n_name = 'FRANCE' and n2.n_name = 'GERMANY')
             or (n1.n_name = 'GERMANY' and n2.n_name = 'FRANCE'))
        and l_shipdate between date '1995-01-01' and date '1996-12-31'
) as shipping
group by
    supp_nation,
    cust_nation,
    l_year
order by
    supp_nation,
    cust_nation,
    l_year
""",
    ),
    2: StaticSQLVariant(
        2,
        "HAVING clause: Use HAVING for volume filtering",
        """\

select
    supp_nation,
    cust_nation,
    l_year,
    sum(volume) as revenue
from (
    select
        n1.n_name as supp_nation,
        n2.n_name as cust_nation,
        extract(year from l_shipdate) as l_year,
        l_extendedprice * (1 - l_discount) as volume
    from
        supplier, lineitem, orders, customer, nation n1, nation n2
    where
        s_suppkey = l_suppkey
        and o_orderkey = l_orderkey
        and c_custkey = o_custkey
        and s_nationkey = n1.n_nationkey
        and c_nationkey = n2.n_nationkey
        and ((n1.n_name = 'FRANCE' and n2.n_name = 'GERMANY')
             or (n1.n_name = 'GERMANY' and n2.n_name = 'FRANCE'))
        and l_shipdate between date '1995-01-01' and date '1996-12-31'
) as shipping
group by
    supp_nation,
    cust_nation,
    l_year
having
    sum(volume) > 0
order by
    supp_nation,
    cust_nation,
    l_year
""",
    ),
    3: StaticSQLVariant(
        3,
        "Explicit JOINs: Convert all implicit joins to explicit JOIN syntax",
        """\

select
    supp_nation,
    cust_nation,
    l_year,
    sum(volume) as revenue
from (
    select
        n1.n_name as supp_nation,
        n2.n_name as cust_nation,
        extract(year from l_shipdate) as l_year,
        l_extendedprice * (1 - l_discount) as volume
    from
        supplier s
        inner join lineitem l on s.s_suppkey = l.l_suppkey
        inner join orders o on l.l_orderkey = o.o_orderkey
        inner join customer c on o.o_custkey = c.c_custkey
        inner join nation n1 on s.s_nationkey = n1.n_nationkey
        inner join nation n2 on c.c_nationkey = n2.n_nationkey
    where
        ((n1.n_name = 'FRANCE' and n2.n_name = 'GERMANY')
         or (n1.n_name = 'GERMANY' and n2.n_name = 'FRANCE'))
        and l.l_shipdate between date '1995-01-01' and date '1996-12-31'
) as shipping
group by
    supp_nation,
    cust_nation,
    l_year
order by
    supp_nation,
    cust_nation,
    l_year
""",
    ),
    4: StaticSQLVariant(
        4,
        "UNION ALL: Split by nation combinations",
        """\

select
    supp_nation,
    cust_nation,
    l_year,
    sum(volume) as revenue
from (
    select
        n1.n_name as supp_nation,
        n2.n_name as cust_nation,
        extract(year from l_shipdate) as l_year,
        l_extendedprice * (1 - l_discount) as volume
    from
        supplier, lineitem, orders, customer, nation n1, nation n2
    where
        s_suppkey = l_suppkey
        and o_orderkey = l_orderkey
        and c_custkey = o_custkey
        and s_nationkey = n1.n_nationkey
        and c_nationkey = n2.n_nationkey
        and n1.n_name = 'FRANCE' and n2.n_name = 'GERMANY'
        and l_shipdate between date '1995-01-01' and date '1996-12-31'

    union all

    select
        n1.n_name as supp_nation,
        n2.n_name as cust_nation,
        extract(year from l_shipdate) as l_year,
        l_extendedprice * (1 - l_discount) as volume
    from
        supplier, lineitem, orders, customer, nation n1, nation n2
    where
        s_suppkey = l_suppkey
        and o_orderkey = l_orderkey
        and c_custkey = o_custkey
        and s_nationkey = n1.n_nationkey
        and c_nationkey = n2.n_nationkey
        and n1.n_name = 'GERMANY' and n2.n_name = 'FRANCE'
        and l_shipdate between date '1995-01-01' and date '1996-12-31'
) as shipping
group by
    supp_nation,
    cust_nation,
    l_year
order by
    supp_nation,
    cust_nation,
    l_year
""",
    ),
    5: StaticSQLVariant(
        5,
        "Common Table Expression: Use CTEs to break down complex logic",
        """\

with qualified_suppliers as (
    select s_suppkey, n_name as supp_nation
    from supplier, nation
    where s_nationkey = n_nationkey
      and n_name in ('FRANCE', 'GERMANY')
),
qualified_customers as (
    select c_custkey, n_name as cust_nation
    from customer, nation
    where c_nationkey = n_nationkey
      and n_name in ('FRANCE', 'GERMANY')
),
qualified_orders as (
    select o_orderkey, o_custkey, extract(year from l_shipdate) as l_year
    from orders, lineitem
    where o_orderkey = l_orderkey
      and l_shipdate between date '1995-01-01' and date '1996-12-31'
),
shipping_base as (
    select
        qs.supp_nation,
        qc.cust_nation,
        qo.l_year,
        l.l_extendedprice * (1 - l.l_discount) as volume
    from
        qualified_suppliers qs
        join lineitem l on qs.s_suppkey = l.l_suppkey
        join qualified_orders qo on l.l_orderkey = qo.o_orderkey
        join qualified_customers qc on qo.o_custkey = qc.c_custkey
    where
        ((qs.supp_nation = 'FRANCE' and qc.cust_nation = 'GERMANY')
         or (qs.supp_nation = 'GERMANY' and qc.cust_nation = 'FRANCE'))
)
select
    supp_nation,
    cust_nation,
    l_year,
    sum(volume) as revenue
from shipping_base
group by
    supp_nation,
    cust_nation,
    l_year
order by
    supp_nation,
    cust_nation,
    l_year
""",
    ),
    6: StaticSQLVariant(
        6,
        "Nested derived tables: Multiple levels of derived tables",
        """\

select
    supp_nation,
    cust_nation,
    l_year,
    total_volume
from (
    select
        supp_nation,
        cust_nation,
        l_year,
        sum(volume) as total_volume
    from (
        select
            shipping_info.supp_nation,
            shipping_info.cust_nation,
            shipping_info.l_year,
            shipping_info.volume
        from (
            select
                n1.n_name as supp_nation,
                n2.n_name as cust_nation,
                extract(year from l_shipdate) as l_year,
                l_extendedprice * (1 - l_discount) as volume
            from
                supplier, lineitem, orders, customer, nation n1, nation n2
            where
                s_suppkey = l_suppkey
                and o_orderkey = l_orderkey
                and c_custkey = o_custkey
                and s_nationkey = n1.n_nationkey
                and c_nationkey = n2.n_nationkey
                and ((n1.n_name = 'FRANCE' and n2.n_name = 'GERMANY')
                     or (n1.n_name = 'GERMANY' and n2.n_name = 'FRANCE'))
                and l_shipdate between date '1995-01-01' and date '1996-12-31'
        ) as shipping_info
    ) as shipping_details
    group by
        supp_nation,
        cust_nation,
        l_year
) as shipping_summary
order by
    supp_nation,
    cust_nation,
    l_year
""",
    ),
    7: StaticSQLVariant(
        7,
        "FILTER clause: Use FILTER for conditional aggregation (OLAP)",
        """\

select
    supp_nation,
    cust_nation,
    l_year,
    sum(l_extendedprice * (1 - l_discount))
        filter (where l_shipdate between date '1995-01-01' and date '1996-12-31') as revenue
from (
    select
        n1.n_name as supp_nation,
        n2.n_name as cust_nation,
        extract(year from l_shipdate) as l_year,
        l_extendedprice,
        l_discount,
        l_shipdate
    from
        supplier, lineitem, orders, customer, nation n1, nation n2
    where
        s_suppkey = l_suppkey
        and o_orderkey = l_orderkey
        and c_custkey = o_custkey
        and s_nationkey = n1.n_nationkey
        and c_nationkey = n2.n_nationkey
        and ((n1.n_name = 'FRANCE' and n2.n_name = 'GERMANY')
             or (n1.n_name = 'GERMANY' and n2.n_name = 'FRANCE'))
) as shipping
group by
    supp_nation,
    cust_nation,
    l_year
having
    revenue is not null
order by
    supp_nation,
    cust_nation,
    l_year
""",
    ),
    8: StaticSQLVariant(
        8,
        "EXISTS pattern: Use EXISTS for nation filtering",
        """\

select
    supp_nation,
    cust_nation,
    l_year,
    sum(volume) as revenue
from (
    select
        n1.n_name as supp_nation,
        n2.n_name as cust_nation,
        extract(year from l_shipdate) as l_year,
        l_extendedprice * (1 - l_discount) as volume
    from
        supplier, lineitem, orders, customer, nation n1, nation n2
    where
        s_suppkey = l_suppkey
        and o_orderkey = l_orderkey
        and c_custkey = o_custkey
        and s_nationkey = n1.n_nationkey
        and c_nationkey = n2.n_nationkey
        and l_shipdate between date '1995-01-01' and date '1996-12-31'
        and exists (
            select 1
            where (n1.n_name = 'FRANCE' and n2.n_name = 'GERMANY')
               or (n1.n_name = 'GERMANY' and n2.n_name = 'FRANCE')
        )
) as shipping
group by
    supp_nation,
    cust_nation,
    l_year
order by
    supp_nation,
    cust_nation,
    l_year
""",
    ),
    9: StaticSQLVariant(
        9,
        "Window functions: Use window functions for revenue analysis (OLAP)",
        """\

select distinct
    supp_nation,
    cust_nation,
    l_year,
    sum(volume) over (partition by supp_nation, cust_nation, l_year) as revenue,
    rank() over (partition by l_year order by sum(volume) over (partition by supp_nation, cust_nation, l_year) desc) as revenue_rank
from (
    select
        n1.n_name as supp_nation,
        n2.n_name as cust_nation,
        extract(year from l_shipdate) as l_year,
        l_extendedprice * (1 - l_discount) as volume
    from
        supplier, lineitem, orders, customer, nation n1, nation n2
    where
        s_suppkey = l_suppkey
        and o_orderkey = l_orderkey
        and c_custkey = o_custkey
        and s_nationkey = n1.n_nationkey
        and c_nationkey = n2.n_nationkey
        and ((n1.n_name = 'FRANCE' and n2.n_name = 'GERMANY')
             or (n1.n_name = 'GERMANY' and n2.n_name = 'FRANCE'))
        and l_shipdate between date '1995-01-01' and date '1996-12-31'
) as shipping
order by
    supp_nation,
    cust_nation,
    l_year
""",
    ),
    10: StaticSQLVariant(
        10,
        "CASE expressions: Add extensive CASE logic for conditional processing",
        """\

select
    case
        when supp_nation = 'FRANCE' then 'FRANCE'
        when supp_nation = 'GERMANY' then 'GERMANY'
        else supp_nation
    end as supp_nation,
    case
        when cust_nation = 'FRANCE' then 'FRANCE'
        when cust_nation = 'GERMANY' then 'GERMANY'
        else cust_nation
    end as cust_nation,
    l_year,
    sum(
        case
            when volume > 0 then volume
            else 0
        end
    ) as revenue
from (
    select
        case
            when n1.n_name in ('FRANCE', 'GERMANY') then n1.n_name
            else 'OTHER'
        end as supp_nation,
        case
            when n2.n_name in ('FRANCE', 'GERMANY') then n2.n_name
            else 'OTHER'
        end as cust_nation,
        extract(year from l_shipdate) as l_year,
        case
            when l_discount between 0.05 and 0.07 then l_extendedprice * (1 - l_discount)
            when l_discount < 0.05 then l_extendedprice * (1 - l_discount)
            when l_discount > 0.07 then l_extendedprice * (1 - l_discount)
            else l_extendedprice * (1 - l_discount)
        end as volume
    from
        supplier, lineitem, orders, customer, nation n1, nation n2
    where
        s_suppkey = l_suppkey
        and o_orderkey = l_orderkey
        and c_custkey = o_custkey
        and s_nationkey = n1.n_nationkey
        and c_nationkey = n2.n_nationkey
        and case
            when (n1.n_name = 'FRANCE' and n2.n_name = 'GERMANY') then 1
            when (n1.n_name = 'GERMANY' and n2.n_name = 'FRANCE') then 1
            else 0
        end = 1
        and case
            when l_shipdate >= date '1995-01-01' then 1
            else 0
        end = 1
        and case
            when l_shipdate <= date '1996-12-31' then 1
            else 0
        end = 1
) as shipping
group by
    supp_nation,
    cust_nation,
    l_year
order by
    supp_nation,
    cust_nation,
    l_year
""",
    ),
}

__all__ = ["VARIANTS"]



"""Variant definitions for Query 13."""

from __future__ import annotations

from benchbox.core.tpchavoc.variant_base import StaticSQLVariant

VARIANTS = {
    1: StaticSQLVariant(
        1,
        "Scalar subqueries: Use scalar subqueries for order counting",
        """\

select
    c_count,
    count(*) as custdist
from (
    select
        c_custkey,
        (select count(o2.o_orderkey)
         from orders o2
         where o2.o_custkey = c.c_custkey
           and o2.o_comment not like '%special%requests%') as c_count
    from customer c
) as c_orders
group by c_count
order by custdist desc, c_count desc
""",
    ),
    2: StaticSQLVariant(
        2,
        "HAVING clause: Use HAVING for distribution filtering",
        """\

select
    c_count,
    count(*) as custdist
from (
    select
        c_custkey,
        count(o_orderkey) as c_count
    from
        customer left outer join orders on
            c_custkey = o_custkey
            and o_comment not like '%special%requests%'
    group by
        c_custkey
) as c_orders
group by
    c_count
having
    count(*) > 0
order by
    custdist desc,
    c_count desc
""",
    ),
    3: StaticSQLVariant(
        3,
        "Explicit JOINs: Convert LEFT JOIN to explicit syntax",
        """\

select
    c_count,
    count(*) as custdist
from (
    select
        c.c_custkey,
        count(o.o_orderkey) as c_count
    from
        customer c
        left outer join orders o on c.c_custkey = o.o_custkey
                                 and o.o_comment not like '%special%requests%'
    group by
        c.c_custkey
) as c_orders (c_custkey, c_count)
group by
    c_count
order by
    custdist desc,
    c_count desc
""",
    ),
    4: StaticSQLVariant(
        4,
        "UNION ALL: Split customer analysis by segments",
        """\

select
    c_count,
    sum(custdist) as custdist
from (
    select c_count, count(*) as custdist
    from (
        select c_custkey, count(o_orderkey) as c_count
        from customer left outer join orders on c_custkey = o_custkey
             and o_comment not like '%special%requests%'
        where c_custkey % 2 = 0
        group by c_custkey
    ) c_orders_even
    group by c_count

    union all

    select c_count, count(*) as custdist
    from (
        select c_custkey, count(o_orderkey) as c_count
        from customer left outer join orders on c_custkey = o_custkey
             and o_comment not like '%special%requests%'
        where c_custkey % 2 = 1
        group by c_custkey
    ) c_orders_odd
    group by c_count
) combined
group by c_count
order by custdist desc, c_count desc
""",
    ),
    5: StaticSQLVariant(
        5,
        "Common Table Expression: Use CTEs to break down double aggregation",
        """\

with filtered_orders as (
    select o_orderkey, o_custkey
    from orders
    where o_comment not like '%special%requests%'
),
customer_order_counts as (
    select
        c.c_custkey,
        count(fo.o_orderkey) as c_count
    from
        customer c
        left outer join filtered_orders fo on c.c_custkey = fo.o_custkey
    group by
        c.c_custkey
),
distribution_analysis as (
    select c_count, count(*) as custdist
    from customer_order_counts
    group by c_count
)
select c_count, custdist
from distribution_analysis
order by custdist desc, c_count desc
""",
    ),
    6: StaticSQLVariant(
        6,
        "Derived table: Multiple levels of derived tables",
        """\

select
    order_count,
    customer_distribution
from (
    select
        c_count as order_count,
        count(*) as customer_distribution
    from (
        select
            customer_orders.c_custkey,
            customer_orders.c_count
        from (
            select
                c_custkey,
                count(o_orderkey) as c_count
            from
                customer left outer join orders on
                    c_custkey = o_custkey
                    and o_comment not like '%special%requests%'
            group by
                c_custkey
        ) as customer_orders
    ) as customer_summary
    group by
        c_count
) as distribution_summary
order by
    customer_distribution desc,
    order_count desc
""",
    ),
    7: StaticSQLVariant(
        7,
        "FILTER clause: Use FILTER for conditional counting (OLAP)",
        """\

select
    c_count,
    count(*) as custdist
from (
    select
        c_custkey,
        count(o_orderkey) filter (where o_comment not like '%special%requests%') as c_count
    from
        customer left outer join orders on c_custkey = o_custkey
    group by
        c_custkey
) as c_orders
group by
    c_count
order by
    custdist desc,
    c_count desc
""",
    ),
    8: StaticSQLVariant(
        8,
        "EXISTS pattern: Use EXISTS for order filtering",
        """\

select
    c_count,
    count(*) as custdist
from (
    select
        c_custkey,
        (select count(*)
         from orders o
         where o.o_custkey = c.c_custkey
           and not exists (
               select 1
               where o.o_comment like '%special%requests%'
           )) as c_count
    from customer c
) as c_orders
group by
    c_count
order by
    custdist desc,
    c_count desc
""",
    ),
    9: StaticSQLVariant(
        9,
        "Window functions: Use window functions for distribution analysis (OLAP)",
        """\

select distinct
    c_count,
    count(*) over (partition by c_count) as custdist,
    rank() over (order by count(*) over (partition by c_count) desc, c_count desc) as dist_rank
from (
    select
        c_custkey,
        count(o_orderkey) as c_count
    from
        customer left outer join orders on
            c_custkey = o_custkey
            and o_comment not like '%special%requests%'
    group by
        c_custkey
) as c_orders
order by
    custdist desc,
    c_count desc
""",
    ),
    10: StaticSQLVariant(
        10,
        "CASE expressions: Add extensive CASE logic for order classification",
        """\

select
    case
        when c_count = 0 then 0
        when c_count between 1 and 5 then c_count
        when c_count between 6 and 10 then c_count
        when c_count > 10 then c_count
        else c_count
    end as c_count,
    count(*) as custdist
from (
    select
        c_custkey,
        sum(
            case
                when case
                    when o_comment not like '%special%requests%' then 1
                    when o_comment is null then 1
                    else 0
                end = 1 then
                    case
                        when o_orderkey is not null then 1
                        else 0
                    end
                else 0
            end
        ) as c_count
    from
        customer left outer join orders on c_custkey = o_custkey
    group by
        c_custkey
) as c_orders
group by
    c_count
order by
    custdist desc,
    c_count desc
""",
    ),
}

__all__ = ["VARIANTS"]



"""
SQLite query plan parser.

Parses SQLite's EXPLAIN QUERY PLAN output into QueryPlanDAG structure.
SQLite uses a simple indented text format showing the query execution plan.

Example EXPLAIN QUERY PLAN output:
```
QUERY PLAN
|--SCAN TABLE orders
`--USE TEMP B-TREE FOR ORDER BY
```

Or with joins:
```
QUERY PLAN
|--SCAN TABLE customer
`--SEARCH TABLE orders USING INDEX idx_customer (c_custkey=?)
```
"""

from __future__ import annotations

import logging
import re
from typing import Any

from benchbox.core.query_plans.parsers.base import QueryPlanParser
from benchbox.core.results.query_plan_models import (
    LogicalOperator,
    LogicalOperatorType,
    QueryPlanDAG,
)

logger = logging.getLogger(__name__)


class SQLiteQueryPlanParser(QueryPlanParser):
    """Parser for SQLite EXPLAIN QUERY PLAN output."""

    def __init__(self):
        super().__init__("sqlite")

    def _parse_impl(self, query_id: str, explain_output: str) -> QueryPlanDAG:
        """
        Parse SQLite EXPLAIN QUERY PLAN output.

        Args:
            query_id: Query identifier
            explain_output: SQLite EXPLAIN QUERY PLAN output

        Returns:
            QueryPlanDAG

        Raises:
            ValueError: If output cannot be parsed
        """
        if not explain_output or not explain_output.strip():
            raise ValueError("Empty EXPLAIN output")

        # Parse lines
        lines = explain_output.strip().split("\n")

        # Skip header if present
        if lines and "QUERY PLAN" in lines[0]:
            lines = lines[1:]

        if not lines:
            raise ValueError("No plan lines found")

        # Parse operators from text
        operators = self._parse_text_operators(lines)

        if not operators:
            raise ValueError("No operators found in EXPLAIN output")

        # Build operator tree
        root = self._build_operator_tree(operators)

        return QueryPlanDAG(
            query_id=query_id,
            platform=self.platform_name,
            logical_root=root,
            estimated_cost=None,  # SQLite EXPLAIN QUERY PLAN doesn't provide cost estimates
            estimated_rows=None,
            raw_explain_output=explain_output,
        )

    def _parse_text_operators(self, lines: list[str]) -> list[dict[str, Any]]:
        """
        Parse operator lines from SQLite output.

        SQLite uses indentation with |-- and `-- prefixes to show hierarchy.

        Returns:
            List of operator dictionaries
        """
        operators = []

        for line in lines:
            if not line.strip():
                continue

            # Extract indentation level
            level = 0
            stripped = line.lstrip("|`- ")
            indent_chars = line[: len(line) - len(stripped)]
            level = indent_chars.count("|") + indent_chars.count("`")

            # Parse operator details
            op_dict = self._parse_operator_line(stripped, level)
            if op_dict:
                operators.append(op_dict)

        return operators

    def _parse_operator_line(self, line: str, level: int) -> dict[str, Any] | None:
        """
        Parse a single operator line.

        Args:
            line: Stripped operator description
            level: Indentation level

        Returns:
            Operator dictionary or None
        """
        line = line.strip()
        if not line:
            return None

        # Common SQLite patterns:
        # - SCAN TABLE tablename
        # - SEARCH TABLE tablename USING INDEX idx_name
        # - USE TEMP B-TREE FOR ORDER BY
        # - USE TEMP B-TREE FOR GROUP BY
        # - EXECUTE CORRELATED SCALAR SUBQUERY

        return {
            "text": line,
            "level": level,
            "type": self._infer_operator_type(line),
            "details": self._extract_details(line),
        }

    def _infer_operator_type(self, line: str) -> str:
        """Infer operator type from line text."""
        upper = line.upper()

        if "SCAN TABLE" in upper or "SCAN SUBQUERY" in upper:
            return "SCAN"
        elif "SEARCH TABLE" in upper:
            return "INDEX_SCAN"
        elif "ORDER BY" in upper:
            return "SORT"
        elif "GROUP BY" in upper:
            return "AGGREGATE"
        elif "TEMP B-TREE" in upper and "JOIN" not in upper:
            # Temp B-tree without JOIN context is usually for sorting/grouping
            if "ORDER BY" in upper:
                return "SORT"
            elif "GROUP BY" in upper:
                return "AGGREGATE"
            else:
                return "OTHER"
        elif "COMPOUND" in upper or "UNION" in upper:
            return "UNION"
        elif "SUBQUERY" in upper:
            return "SUBQUERY"
        else:
            return "OTHER"

    def _extract_details(self, line: str) -> dict[str, Any]:
        """Extract details from operator line."""
        details: dict[str, Any] = {}

        # Extract table name from SCAN/SEARCH patterns
        match = re.search(r"(?:SCAN|SEARCH) (?:TABLE|SUBQUERY) (\w+)", line, re.IGNORECASE)
        if match:
            details["table_name"] = match.group(1)

        # Extract index name
        match = re.search(r"USING INDEX (\w+)", line, re.IGNORECASE)
        if match:
            details["index_name"] = match.group(1)

        # Extract join type if present
        if "LEFT" in line.upper():
            details["join_type"] = "left"
        elif "RIGHT" in line.upper():
            details["join_type"] = "right"
        elif "OUTER" in line.upper():
            details["join_type"] = "full"
        elif "JOIN" in line.upper():
            details["join_type"] = "inner"

        return details

    def _build_operator_tree(self, operators: list[dict[str, Any]]) -> LogicalOperator:
        """
        Build operator tree from parsed operators.

        SQLite output is hierarchical with indentation, but for simplicity
        we'll build a linear chain for now.

        Args:
            operators: List of parsed operators

        Returns:
            Root LogicalOperator
        """
        if not operators:
            raise ValueError("No operators to build tree from")

        # Build operators bottom-up (reverse order)
        logical_operators = []

        for op_dict in reversed(operators):
            logical_op = self._convert_to_logical_operator(op_dict, logical_operators[-1:] if logical_operators else [])
            logical_operators.append(logical_op)

        # Return the top operator
        return logical_operators[-1] if logical_operators else self._create_fallback_operator()

    def _convert_to_logical_operator(
        self,
        op_dict: dict[str, Any],
        children: list[LogicalOperator],
    ) -> LogicalOperator:
        """
        Convert parsed operator dict to LogicalOperator.

        Args:
            op_dict: Operator dictionary from parsing
            children: Child operators

        Returns:
            LogicalOperator instance
        """
        op_type_str = op_dict["type"]
        details = op_dict["details"]
        text = op_dict["text"]

        # Map to logical operator type
        logical_type = self._map_sqlite_operator(op_type_str)

        # Extract operator-specific information
        kwargs: dict[str, Any] = {}

        if logical_type == LogicalOperatorType.SCAN:
            table_name = details.get("table_name")
            if table_name:
                kwargs["table_name"] = table_name

        elif logical_type == LogicalOperatorType.SORT:
            # Extract sort information from text if available
            kwargs["properties"] = {"temp_btree": "TEMP B-TREE" in text.upper()}

        elif logical_type == LogicalOperatorType.AGGREGATE:
            kwargs["properties"] = {"temp_btree": "TEMP B-TREE" in text.upper()}

        # Create physical operator with SQLite-specific details
        physical_op = self._create_physical_operator(
            op_type_str,
            properties=details.copy(),
            platform_metadata={"text": text, "level": op_dict["level"]},
        )

        return self._create_logical_operator(
            operator_type=logical_type,
            children=children,
            physical_operator=physical_op,
            **kwargs,
        )

    def _map_sqlite_operator(self, sqlite_op: str) -> LogicalOperatorType:
        """Map SQLite operator string to LogicalOperatorType."""
        normalized = sqlite_op.upper()

        if normalized == "SCAN" or normalized == "INDEX_SCAN":
            return LogicalOperatorType.SCAN
        elif normalized == "SORT":
            return LogicalOperatorType.SORT
        elif normalized == "AGGREGATE":
            return LogicalOperatorType.AGGREGATE
        elif normalized == "UNION":
            return LogicalOperatorType.UNION
        elif normalized == "SUBQUERY":
            return LogicalOperatorType.SUBQUERY
        else:
            return LogicalOperatorType.OTHER

    def _create_fallback_operator(self) -> LogicalOperator:
        """Create fallback operator when parsing fails."""
        return self._create_logical_operator(
            operator_type=LogicalOperatorType.OTHER,
            properties={"note": "Fallback operator due to parsing failure"},
        )



"""
Query plan insights and analysis.

Provides automated analysis of query plans including complexity scoring
and anti-pattern detection.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from benchbox.core.results.query_plan_models import (
    JoinType,
    LogicalOperator,
    LogicalOperatorType,
)

if TYPE_CHECKING:
    from benchbox.core.results.query_plan_models import QueryPlanDAG


@dataclass
class PlanComplexityScore:
    """Query plan complexity metrics."""

    total_operators: int = 0
    join_count: int = 0
    join_complexity: int = 0
    aggregation_count: int = 0
    subquery_depth: int = 0
    max_tree_depth: int = 0
    scan_count: int = 0
    filter_count: int = 0
    sort_count: int = 0
    overall_score: int = 0  # 0-100, higher = more complex
    complexity_level: str = "low"  # "low", "medium", "high", "very_high"

    def to_dict(self) -> dict:
        """Convert to dictionary."""
        return {
            "total_operators": self.total_operators,
            "join_count": self.join_count,
            "join_complexity": self.join_complexity,
            "aggregation_count": self.aggregation_count,
            "subquery_depth": self.subquery_depth,
            "max_tree_depth": self.max_tree_depth,
            "scan_count": self.scan_count,
            "filter_count": self.filter_count,
            "sort_count": self.sort_count,
            "overall_score": self.overall_score,
            "complexity_level": self.complexity_level,
        }


@dataclass
class PlanInsight:
    """Single insight or finding about a query plan."""

    category: str  # "warning", "optimization", "info"
    title: str
    description: str
    operator_id: str | None = None
    severity: str = "medium"  # "low", "medium", "high", "critical"

    def to_dict(self) -> dict:
        """Convert to dictionary."""
        return {
            "category": self.category,
            "title": self.title,
            "description": self.description,
            "operator_id": self.operator_id,
            "severity": self.severity,
        }


@dataclass
class PlanAnalysisResult:
    """Complete analysis result for a query plan."""

    query_id: str
    complexity: PlanComplexityScore
    insights: list[PlanInsight] = field(default_factory=list)

    def to_dict(self) -> dict:
        """Convert to dictionary."""
        return {
            "query_id": self.query_id,
            "complexity": self.complexity.to_dict(),
            "insights": [i.to_dict() for i in self.insights],
        }


def _get_join_complexity(op: LogicalOperator) -> int:
    """Return join complexity weight for the given operator."""
    if op.join_type == JoinType.CROSS:
        return 3
    if op.join_type in (JoinType.FULL, JoinType.ANTI):
        return 2
    return 1


def _classify_complexity_level(overall_score: int) -> str:
    """Classify numeric complexity score into a named level."""
    if overall_score < 20:
        return "low"
    if overall_score < 50:
        return "medium"
    if overall_score < 80:
        return "high"
    return "very_high"


def compute_complexity_score(plan: QueryPlanDAG) -> PlanComplexityScore:
    """
    Compute complexity score from plan structure.

    The complexity score helps identify queries that may need optimization
    attention due to their structural complexity.

    Args:
        plan: Query plan to analyze

    Returns:
        PlanComplexityScore with detailed metrics
    """
    score = PlanComplexityScore()

    # Dispatch map: operator type -> score field incrementer
    _counter_map = {
        LogicalOperatorType.AGGREGATE: "aggregation_count",
        LogicalOperatorType.SCAN: "scan_count",
        LogicalOperatorType.FILTER: "filter_count",
        LogicalOperatorType.SORT: "sort_count",
    }

    def analyze_node(op: LogicalOperator, depth: int) -> None:
        score.total_operators += 1
        score.max_tree_depth = max(score.max_tree_depth, depth)

        if op.operator_type == LogicalOperatorType.JOIN:
            score.join_count += 1
            score.join_complexity += _get_join_complexity(op)
        elif op.operator_type == LogicalOperatorType.SUBQUERY:
            score.subquery_depth = max(score.subquery_depth, 1)
        elif op.operator_type in _counter_map:
            field_name = _counter_map[op.operator_type]
            setattr(score, field_name, getattr(score, field_name) + 1)

        for child in op.children:
            analyze_node(child, depth + 1)

    if plan.logical_root:
        analyze_node(plan.logical_root, 0)

    # Compute overall score (0-100)
    score.overall_score = min(
        100,
        (
            score.total_operators * 2
            + score.join_complexity * 8
            + score.aggregation_count * 5
            + score.subquery_depth * 15
            + max(0, score.max_tree_depth - 3) * 3
        ),
    )

    score.complexity_level = _classify_complexity_level(score.overall_score)

    return score


def detect_antipatterns(plan: QueryPlanDAG) -> list[PlanInsight]:
    """
    Detect common anti-patterns in query plan.

    Identifies potential performance issues such as:
    - Cartesian products (cross joins without conditions)
    - Multiple sequential scans on same table
    - Sort operations after aggregations
    - Deep nested joins

    Args:
        plan: Query plan to analyze

    Returns:
        List of detected anti-patterns as PlanInsight objects
    """
    insights: list[PlanInsight] = []

    if not plan.logical_root:
        return insights

    tables_scanned: dict[str, list[str]] = {}  # table_name -> [operator_ids]

    def check_node(op: LogicalOperator, depth: int) -> None:
        # Check for cross joins (potential Cartesian products)
        if op.operator_type == LogicalOperatorType.JOIN:
            if op.join_type == JoinType.CROSS:
                insights.append(
                    PlanInsight(
                        category="warning",
                        title="Cross Join Detected",
                        description="Cross join may produce large intermediate result. "
                        "Ensure this is intentional and not a missing join condition.",
                        operator_id=op.operator_id,
                        severity="high",
                    )
                )

            # Check for join without apparent condition
            if not op.join_conditions and op.join_type != JoinType.CROSS:
                insights.append(
                    PlanInsight(
                        category="info",
                        title="Join Without Visible Condition",
                        description="Join operator has no visible join condition. "
                        "This may be normal if condition is handled by filter operator.",
                        operator_id=op.operator_id,
                        severity="low",
                    )
                )

        # Track table scans
        if op.operator_type == LogicalOperatorType.SCAN and op.table_name:
            if op.table_name not in tables_scanned:
                tables_scanned[op.table_name] = []
            tables_scanned[op.table_name].append(op.operator_id)

        # Check for deep nesting
        if depth > 8:
            insights.append(
                PlanInsight(
                    category="info",
                    title="Deep Plan Nesting",
                    description=f"Plan has depth of {depth}+ levels. Very deep plans may indicate complex subqueries.",
                    operator_id=op.operator_id,
                    severity="low",
                )
            )

        for child in op.children:
            check_node(child, depth + 1)

    check_node(plan.logical_root, 0)

    # Check for multiple scans on same table
    for table_name, scan_ops in tables_scanned.items():
        if len(scan_ops) > 1:
            insights.append(
                PlanInsight(
                    category="optimization",
                    title=f"Multiple Scans on {table_name}",
                    description=f"Table {table_name} is scanned {len(scan_ops)} times. "
                    "Consider rewriting query to scan table once.",
                    operator_id=scan_ops[0],
                    severity="medium",
                )
            )

    return insights


def analyze_plan(plan: QueryPlanDAG) -> PlanAnalysisResult:
    """
    Perform complete analysis of a query plan.

    Combines complexity scoring and anti-pattern detection into
    a comprehensive analysis result.

    Args:
        plan: Query plan to analyze

    Returns:
        PlanAnalysisResult with complexity and insights
    """
    complexity = compute_complexity_score(plan)
    insights = detect_antipatterns(plan)

    # Add complexity-based insights
    if complexity.complexity_level == "very_high":
        insights.append(
            PlanInsight(
                category="warning",
                title="Very High Complexity",
                description=f"Query has complexity score of {complexity.overall_score}/100. "
                "Consider breaking into smaller queries or reviewing join strategy.",
                severity="high",
            )
        )

    if complexity.join_count > 5:
        insights.append(
            PlanInsight(
                category="info",
                title="Many Joins",
                description=f"Query has {complexity.join_count} join operations. Verify join order is optimal.",
                severity="medium",
            )
        )

    return PlanAnalysisResult(
        query_id=plan.query_id,
        complexity=complexity,
        insights=insights,
    )



"""NYC Taxi OLAP benchmark queries.

Implements representative OLAP analytics queries:
- Temporal aggregations (hourly, daily, monthly)
- Geographic patterns (zone-level analytics)
- Financial analytics (revenue, tips, fares)
- Multi-dimensional analysis

Query sources:
- Todd Schneider's nyc-taxi-data repository
- ClickHouse official documentation
- DuckDB taxi dataset benchmarks
- Mark Litwintschik's database performance analyses

Copyright 2026 Joe Harris / BenchBox Project

Licensed under the MIT License. See LICENSE file in the project root for details.
"""

from datetime import datetime, timedelta
from typing import Any, Optional

import numpy as np

# Query definitions - 25 representative OLAP queries
QUERIES = {
    # ==== Temporal Aggregations ====
    "trips-per-hour": {
        "id": "1",
        "name": "Trips per Hour",
        "description": "Count trips by hour of day",
        "category": "temporal",
        "sql": """
            SELECT
                EXTRACT(HOUR FROM pickup_datetime) as hour,
                COUNT(*) as trip_count
            FROM trips
            WHERE pickup_datetime >= '{start_date}'
              AND pickup_datetime < '{end_date}'
            GROUP BY EXTRACT(HOUR FROM pickup_datetime)
            ORDER BY hour
        """,
        "params": {"duration_days": 30},
    },
    "trips-per-day": {
        "id": "2",
        "name": "Trips per Day",
        "description": "Daily trip counts",
        "category": "temporal",
        "sql": """
            SELECT
                DATE_TRUNC('day', pickup_datetime) as day,
                COUNT(*) as trip_count
            FROM trips
            WHERE pickup_datetime >= '{start_date}'
              AND pickup_datetime < '{end_date}'
            GROUP BY DATE_TRUNC('day', pickup_datetime)
            ORDER BY day
        """,
        "params": {"duration_days": 365},
    },
    "trips-per-month": {
        "id": "3",
        "name": "Trips per Month",
        "description": "Monthly trip counts over time",
        "category": "temporal",
        "sql": """
            SELECT
                DATE_TRUNC('month', pickup_datetime) as month,
                COUNT(*) as trip_count,
                SUM(total_amount) as total_revenue
            FROM trips
            WHERE pickup_datetime >= '{start_date}'
              AND pickup_datetime < '{end_date}'
            GROUP BY DATE_TRUNC('month', pickup_datetime)
            ORDER BY month
        """,
        "params": {"duration_days": 365},
    },
    "trips-by-day-of-week": {
        "id": "4",
        "name": "Trips by Day of Week",
        "description": "Trip patterns by weekday",
        "category": "temporal",
        "sql": """
            SELECT
                EXTRACT(DOW FROM pickup_datetime) as day_of_week,
                COUNT(*) as trip_count,
                AVG(trip_distance) as avg_distance,
                AVG(total_amount) as avg_fare
            FROM trips
            WHERE pickup_datetime >= '{start_date}'
              AND pickup_datetime < '{end_date}'
            GROUP BY EXTRACT(DOW FROM pickup_datetime)
            ORDER BY day_of_week
        """,
        "params": {"duration_days": 90},
    },
    # ==== Geographic Analytics ====
    "top-pickup-zones": {
        "id": "5",
        "name": "Top Pickup Zones",
        "description": "Most popular pickup locations",
        "category": "geographic",
        "sql": """
            SELECT
                t.pickup_location_id,
                z.zone,
                z.borough,
                COUNT(*) as trip_count
            FROM trips t
            LEFT JOIN taxi_zones z ON t.pickup_location_id = z.location_id
            WHERE t.pickup_datetime >= '{start_date}'
              AND t.pickup_datetime < '{end_date}'
            GROUP BY t.pickup_location_id, z.zone, z.borough
            ORDER BY trip_count DESC
            LIMIT 20
        """,
        "params": {"duration_days": 30},
    },
    "top-dropoff-zones": {
        "id": "6",
        "name": "Top Dropoff Zones",
        "description": "Most popular dropoff locations",
        "category": "geographic",
        "sql": """
            SELECT
                t.dropoff_location_id,
                z.zone,
                z.borough,
                COUNT(*) as trip_count
            FROM trips t
            LEFT JOIN taxi_zones z ON t.dropoff_location_id = z.location_id
            WHERE t.pickup_datetime >= '{start_date}'
              AND t.pickup_datetime < '{end_date}'
            GROUP BY t.dropoff_location_id, z.zone, z.borough
            ORDER BY trip_count DESC
            LIMIT 20
        """,
        "params": {"duration_days": 30},
    },
    "top-routes": {
        "id": "7",
        "name": "Top Routes",
        "description": "Most popular pickup-dropoff pairs",
        "category": "geographic",
        "sql": """
            SELECT
                t.pickup_location_id,
                pz.zone as pickup_zone,
                t.dropoff_location_id,
                dz.zone as dropoff_zone,
                COUNT(*) as trip_count,
                AVG(t.trip_distance) as avg_distance,
                AVG(t.total_amount) as avg_fare
            FROM trips t
            LEFT JOIN taxi_zones pz ON t.pickup_location_id = pz.location_id
            LEFT JOIN taxi_zones dz ON t.dropoff_location_id = dz.location_id
            WHERE t.pickup_datetime >= '{start_date}'
              AND t.pickup_datetime < '{end_date}'
            GROUP BY t.pickup_location_id, pz.zone, t.dropoff_location_id, dz.zone
            ORDER BY trip_count DESC
            LIMIT 50
        """,
        "params": {"duration_days": 30},
    },
    "borough-summary": {
        "id": "8",
        "name": "Borough Summary",
        "description": "Trip statistics by borough",
        "category": "geographic",
        "sql": """
            SELECT
                z.borough,
                COUNT(*) as trip_count,
                SUM(t.total_amount) as total_revenue,
                AVG(t.trip_distance) as avg_distance,
                AVG(t.tip_amount) as avg_tip
            FROM trips t
            LEFT JOIN taxi_zones z ON t.pickup_location_id = z.location_id
            WHERE t.pickup_datetime >= '{start_date}'
              AND t.pickup_datetime < '{end_date}'
              AND z.borough IS NOT NULL
            GROUP BY z.borough
            ORDER BY trip_count DESC
        """,
        "params": {"duration_days": 90},
    },
    # ==== Financial Analytics ====
    "revenue-by-payment-type": {
        "id": "9",
        "name": "Revenue by Payment Type",
        "description": "Revenue breakdown by payment method",
        "category": "financial",
        "sql": """
            SELECT
                payment_type,
                COUNT(*) as trip_count,
                SUM(total_amount) as total_revenue,
                SUM(tip_amount) as total_tips,
                AVG(tip_amount / NULLIF(fare_amount, 0)) * 100 as avg_tip_percentage
            FROM trips
            WHERE pickup_datetime >= '{start_date}'
              AND pickup_datetime < '{end_date}'
              AND fare_amount > 0
            GROUP BY payment_type
            ORDER BY total_revenue DESC
        """,
        "params": {"duration_days": 30},
    },
    "fare-distribution": {
        "id": "10",
        "name": "Fare Distribution",
        "description": "Distribution of fare amounts",
        "category": "financial",
        "sql": """
            SELECT
                FLOOR(fare_amount / 5) * 5 as fare_bucket,
                COUNT(*) as trip_count,
                AVG(trip_distance) as avg_distance,
                AVG(tip_amount) as avg_tip
            FROM trips
            WHERE pickup_datetime >= '{start_date}'
              AND pickup_datetime < '{end_date}'
              AND fare_amount BETWEEN 0 AND 100
            GROUP BY FLOOR(fare_amount / 5) * 5
            ORDER BY fare_bucket
        """,
        "params": {"duration_days": 30},
    },
    "tip-analysis": {
        "id": "11",
        "name": "Tip Analysis",
        "description": "Tip patterns by time and location",
        "category": "financial",
        "sql": """
            SELECT
                EXTRACT(HOUR FROM pickup_datetime) as hour,
                COUNT(*) as trip_count,
                AVG(tip_amount) as avg_tip,
                AVG(CASE WHEN fare_amount > 0 THEN tip_amount / fare_amount * 100 END) as avg_tip_pct
            FROM trips
            WHERE pickup_datetime >= '{start_date}'
              AND pickup_datetime < '{end_date}'
              AND payment_type = 1
            GROUP BY EXTRACT(HOUR FROM pickup_datetime)
            ORDER BY hour
        """,
        "params": {"duration_days": 30},
    },
    "surcharge-revenue": {
        "id": "12",
        "name": "Surcharge Revenue",
        "description": "Revenue from surcharges and extras",
        "category": "financial",
        "sql": """
            SELECT
                DATE_TRUNC('month', pickup_datetime) as month,
                SUM(extra) as extra_revenue,
                SUM(mta_tax) as mta_tax_revenue,
                SUM(improvement_surcharge) as improvement_revenue,
                SUM(congestion_surcharge) as congestion_revenue,
                SUM(tolls_amount) as tolls_revenue
            FROM trips
            WHERE pickup_datetime >= '{start_date}'
              AND pickup_datetime < '{end_date}'
            GROUP BY DATE_TRUNC('month', pickup_datetime)
            ORDER BY month
        """,
        "params": {"duration_days": 365},
    },
    # ==== Trip Characteristics ====
    "distance-distribution": {
        "id": "13",
        "name": "Distance Distribution",
        "description": "Distribution of trip distances",
        "category": "characteristics",
        "sql": """
            SELECT
                FLOOR(trip_distance) as distance_miles,
                COUNT(*) as trip_count,
                AVG(total_amount) as avg_fare
            FROM trips
            WHERE pickup_datetime >= '{start_date}'
              AND pickup_datetime < '{end_date}'
              AND trip_distance BETWEEN 0 AND 30
            GROUP BY FLOOR(trip_distance)
            ORDER BY distance_miles
        """,
        "params": {"duration_days": 30},
    },
    "passenger-count-analysis": {
        "id": "14",
        "name": "Passenger Count Analysis",
        "description": "Trip patterns by passenger count",
        "category": "characteristics",
        "sql": """
            SELECT
                passenger_count,
                COUNT(*) as trip_count,
                AVG(trip_distance) as avg_distance,
                AVG(total_amount) as avg_fare,
                AVG(total_amount / NULLIF(trip_distance, 0)) as avg_fare_per_mile
            FROM trips
            WHERE pickup_datetime >= '{start_date}'
              AND pickup_datetime < '{end_date}'
              AND passenger_count BETWEEN 1 AND 6
            GROUP BY passenger_count
            ORDER BY passenger_count
        """,
        "params": {"duration_days": 30},
    },
    "trip-duration-analysis": {
        "id": "15",
        "name": "Trip Duration Analysis",
        "description": "Trip duration patterns",
        "category": "characteristics",
        "sql": """
            SELECT
                FLOOR(EXTRACT(EPOCH FROM (dropoff_datetime - pickup_datetime)) / 60 / 5) * 5 as duration_bucket_min,
                COUNT(*) as trip_count,
                AVG(trip_distance) as avg_distance,
                AVG(total_amount) as avg_fare
            FROM trips
            WHERE pickup_datetime >= '{start_date}'
              AND pickup_datetime < '{end_date}'
              AND dropoff_datetime > pickup_datetime
              AND EXTRACT(EPOCH FROM (dropoff_datetime - pickup_datetime)) BETWEEN 60 AND 7200
            GROUP BY FLOOR(EXTRACT(EPOCH FROM (dropoff_datetime - pickup_datetime)) / 60 / 5) * 5
            ORDER BY duration_bucket_min
        """,
        "params": {"duration_days": 30},
    },
    # ==== Rate Code Analysis ====
    "rate-code-summary": {
        "id": "16",
        "name": "Rate Code Summary",
        "description": "Trip statistics by rate code",
        "category": "rates",
        "sql": """
            SELECT
                rate_code_id,
                COUNT(*) as trip_count,
                AVG(trip_distance) as avg_distance,
                AVG(total_amount) as avg_fare,
                SUM(total_amount) as total_revenue
            FROM trips
            WHERE pickup_datetime >= '{start_date}'
              AND pickup_datetime < '{end_date}'
            GROUP BY rate_code_id
            ORDER BY trip_count DESC
        """,
        "params": {"duration_days": 90},
    },
    "airport-trips": {
        "id": "17",
        "name": "Airport Trips",
        "description": "Analysis of airport trips (rate codes 2, 3)",
        "category": "rates",
        "sql": """
            SELECT
                rate_code_id,
                z.zone as pickup_zone,
                COUNT(*) as trip_count,
                AVG(trip_distance) as avg_distance,
                AVG(total_amount) as avg_fare,
                AVG(tip_amount) as avg_tip
            FROM trips t
            LEFT JOIN taxi_zones z ON t.pickup_location_id = z.location_id
            WHERE t.pickup_datetime >= '{start_date}'
              AND t.pickup_datetime < '{end_date}'
              AND t.rate_code_id IN (2, 3)
            GROUP BY rate_code_id, z.zone
            ORDER BY trip_count DESC
            LIMIT 20
        """,
        "params": {"duration_days": 90},
    },
    # ==== Vendor Analysis ====
    "vendor-comparison": {
        "id": "18",
        "name": "Vendor Comparison",
        "description": "Compare vendors by various metrics",
        "category": "vendor",
        "sql": """
            SELECT
                vendor_id,
                COUNT(*) as trip_count,
                AVG(trip_distance) as avg_distance,
                AVG(total_amount) as avg_fare,
                AVG(tip_amount) as avg_tip,
                SUM(total_amount) as total_revenue
            FROM trips
            WHERE pickup_datetime >= '{start_date}'
              AND pickup_datetime < '{end_date}'
            GROUP BY vendor_id
            ORDER BY trip_count DESC
        """,
        "params": {"duration_days": 30},
    },
    # ==== Complex Analytics ====
    "hourly-zone-heatmap": {
        "id": "19",
        "name": "Hourly Zone Heatmap",
        "description": "Trip density by hour and zone",
        "category": "complex",
        "sql": """
            SELECT
                EXTRACT(HOUR FROM pickup_datetime) as hour,
                pickup_location_id,
                COUNT(*) as trip_count
            FROM trips
            WHERE pickup_datetime >= '{start_date}'
              AND pickup_datetime < '{end_date}'
            GROUP BY EXTRACT(HOUR FROM pickup_datetime), pickup_location_id
            ORDER BY hour, trip_count DESC
        """,
        "params": {"duration_days": 7},
    },
    "weekday-weekend-comparison": {
        "id": "20",
        "name": "Weekday vs Weekend",
        "description": "Compare weekday and weekend patterns",
        "category": "complex",
        "sql": """
            SELECT
                CASE WHEN EXTRACT(DOW FROM pickup_datetime) IN (0, 6) THEN 'weekend' ELSE 'weekday' END as day_type,
                EXTRACT(HOUR FROM pickup_datetime) as hour,
                COUNT(*) as trip_count,
                AVG(trip_distance) as avg_distance,
                AVG(total_amount) as avg_fare
            FROM trips
            WHERE pickup_datetime >= '{start_date}'
              AND pickup_datetime < '{end_date}'
            GROUP BY
                CASE WHEN EXTRACT(DOW FROM pickup_datetime) IN (0, 6) THEN 'weekend' ELSE 'weekday' END,
                EXTRACT(HOUR FROM pickup_datetime)
            ORDER BY day_type, hour
        """,
        "params": {"duration_days": 30},
    },
    "rush-hour-analysis": {
        "id": "21",
        "name": "Rush Hour Analysis",
        "description": "Analysis of rush hour patterns",
        "category": "complex",
        "sql": """
            SELECT
                CASE
                    WHEN EXTRACT(HOUR FROM pickup_datetime) BETWEEN 7 AND 9 THEN 'morning_rush'
                    WHEN EXTRACT(HOUR FROM pickup_datetime) BETWEEN 17 AND 19 THEN 'evening_rush'
                    ELSE 'off_peak'
                END as period,
                COUNT(*) as trip_count,
                AVG(trip_distance) as avg_distance,
                AVG(total_amount) as avg_fare,
                AVG(EXTRACT(EPOCH FROM (dropoff_datetime - pickup_datetime)) / 60) as avg_duration_min
            FROM trips
            WHERE pickup_datetime >= '{start_date}'
              AND pickup_datetime < '{end_date}'
              AND dropoff_datetime > pickup_datetime
            GROUP BY
                CASE
                    WHEN EXTRACT(HOUR FROM pickup_datetime) BETWEEN 7 AND 9 THEN 'morning_rush'
                    WHEN EXTRACT(HOUR FROM pickup_datetime) BETWEEN 17 AND 19 THEN 'evening_rush'
                    ELSE 'off_peak'
                END
            ORDER BY trip_count DESC
        """,
        "params": {"duration_days": 30},
    },
    "monthly-year-over-year": {
        "id": "22",
        "name": "Monthly Year-over-Year",
        "description": "Compare monthly patterns across years",
        "category": "complex",
        "sql": """
            SELECT
                EXTRACT(YEAR FROM pickup_datetime) as year,
                EXTRACT(MONTH FROM pickup_datetime) as month,
                COUNT(*) as trip_count,
                SUM(total_amount) as total_revenue,
                AVG(total_amount) as avg_fare
            FROM trips
            WHERE pickup_datetime >= '{start_date}'
              AND pickup_datetime < '{end_date}'
            GROUP BY EXTRACT(YEAR FROM pickup_datetime), EXTRACT(MONTH FROM pickup_datetime)
            ORDER BY year, month
        """,
        "params": {"duration_days": 730},
    },
    # ==== Point Queries ====
    "single-day-summary": {
        "id": "23",
        "name": "Single Day Summary",
        "description": "Summary statistics for a single day",
        "category": "point",
        "sql": """
            SELECT
                COUNT(*) as trip_count,
                SUM(total_amount) as total_revenue,
                AVG(trip_distance) as avg_distance,
                AVG(total_amount) as avg_fare,
                AVG(tip_amount) as avg_tip,
                MIN(pickup_datetime) as first_trip,
                MAX(pickup_datetime) as last_trip
            FROM trips
            WHERE pickup_datetime >= '{start_date}'
              AND pickup_datetime < '{end_date}'
        """,
        "params": {"duration_days": 1},
    },
    "zone-detail": {
        "id": "24",
        "name": "Zone Detail",
        "description": "Detailed statistics for a specific zone",
        "category": "point",
        "sql": """
            SELECT
                z.zone,
                z.borough,
                COUNT(*) as pickup_count,
                AVG(t.trip_distance) as avg_distance,
                AVG(t.total_amount) as avg_fare,
                SUM(t.total_amount) as total_revenue
            FROM trips t
            JOIN taxi_zones z ON t.pickup_location_id = z.location_id
            WHERE t.pickup_datetime >= '{start_date}'
              AND t.pickup_datetime < '{end_date}'
              AND t.pickup_location_id = {zone_id}
            GROUP BY z.zone, z.borough
        """,
        "params": {"duration_days": 30},
    },
    "full-scan-count": {
        "id": "25",
        "name": "Full Scan Count",
        "description": "Count all trips (baseline scan query)",
        "category": "baseline",
        "sql": """
            SELECT COUNT(*) as total_trips
            FROM trips
        """,
        "params": {},
    },
}


class NYCTaxiQueryManager:
    """Manages NYC Taxi benchmark queries."""

    def __init__(
        self,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        seed: Optional[int] = None,
    ) -> None:
        """Initialize query manager.

        Args:
            start_date: Start date for the dataset
            end_date: End date for the dataset
            seed: Random seed for parameter generation
        """
        self.start_date = start_date or datetime(2019, 1, 1)
        self.end_date = end_date or datetime(2019, 12, 31)
        self.rng = np.random.default_rng(seed)

        # Popular zone IDs for point queries
        self.popular_zones = [132, 138, 161, 162, 163, 164, 186, 230, 234, 236, 237, 239, 261, 262, 263]

    def get_query(
        self,
        query_id: str,
        params: Optional[dict[str, Any]] = None,
    ) -> str:
        """Get a query with parameters filled in.

        Args:
            query_id: Query identifier
            params: Optional parameter overrides

        Returns:
            Parameterized query string

        Raises:
            ValueError: If query_id is unknown
        """
        if query_id not in QUERIES:
            raise ValueError(f"Unknown query: {query_id}. Available: {list(QUERIES.keys())}")

        query_def = QUERIES[query_id]
        sql = query_def["sql"].strip()

        # Build parameters
        query_params = self._generate_params(query_def, params)

        # Format the query
        return sql.format(**query_params)

    def _generate_params(
        self,
        query_def: dict[str, Any],
        overrides: Optional[dict[str, Any]] = None,
    ) -> dict[str, Any]:
        """Generate query parameters.

        Args:
            query_def: Query definition
            overrides: Parameter overrides

        Returns:
            Dictionary of parameters
        """
        params = {}
        query_params = query_def.get("params", {})
        overrides = overrides or {}

        # Calculate date range
        duration_days = query_params.get("duration_days", 30)

        # Pick a random start point within the dataset
        dataset_days = (self.end_date - self.start_date).days
        max_offset = max(1, dataset_days - duration_days)
        offset_days = int(self.rng.integers(0, max_offset))

        start = self.start_date + timedelta(days=offset_days)
        end = start + timedelta(days=duration_days)

        params["start_date"] = start.strftime("%Y-%m-%d")
        params["end_date"] = end.strftime("%Y-%m-%d")

        # Random zone for point queries
        params["zone_id"] = overrides.get(
            "zone_id",
            self.popular_zones[int(self.rng.integers(0, len(self.popular_zones)))],
        )

        # Apply overrides
        params.update(overrides)

        return params

    def get_queries(self) -> dict[str, str]:
        """Get all queries with generated parameters.

        Returns:
            Dictionary mapping query IDs to query strings
        """
        return {qid: self.get_query(qid) for qid in QUERIES}

    def get_query_info(self, query_id: str) -> dict[str, Any]:
        """Get query metadata.

        Args:
            query_id: Query identifier

        Returns:
            Query metadata dictionary
        """
        if query_id not in QUERIES:
            raise ValueError(f"Unknown query: {query_id}")
        return QUERIES[query_id]

    def get_queries_by_category(self, category: str) -> list[str]:
        """Get query IDs for a specific category.

        Args:
            category: Query category

        Returns:
            List of query IDs
        """
        return [qid for qid, qdef in QUERIES.items() if qdef.get("category") == category]

    @staticmethod
    def get_categories() -> list[str]:
        """Get all query categories.

        Returns:
            List of unique categories
        """
        return list({str(qdef["category"]) for qdef in QUERIES.values()})

    @staticmethod
    def get_query_count() -> int:
        """Get total number of queries.

        Returns:
            Number of queries
        """
        return len(QUERIES)



"""Sorted Parquet writer for benchmark data.

Reads TBL (pipe-delimited) files generated by TPC tools and writes them
as sorted Parquet files. Sorting is done via PyArrow's compute kernels
to support large datasets without loading everything into memory at once.
"""

from __future__ import annotations

import logging
from pathlib import Path
from typing import Any

import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.csv as csv
import pyarrow.parquet as pq

from benchbox.core.data_organization.config import (
    DataOrganizationConfig,
    SortColumn,
    SortOrder,
)

logger = logging.getLogger(__name__)


class SortedParquetWriter:
    """Writes TBL benchmark data as sorted Parquet files.

    Reads pipe-delimited TBL files (as produced by dbgen/dsdgen), sorts
    them by the configured columns, and writes Parquet output with controlled
    row group sizes for zone map experiments.
    """

    def __init__(self, config: DataOrganizationConfig, schema_registry: dict[str, Any] | None = None) -> None:
        """Initialize the writer.

        Args:
            config: Data organization configuration specifying sort columns,
                row group size, and compression.
            schema_registry: Optional mapping of table names to schema dicts
                (with 'columns' key). Used for column name resolution when
                reading TBL files. If not provided, column names are inferred
                as col_0, col_1, etc.
        """
        self.config = config
        self.schema_registry = schema_registry or {}

    def write_sorted_parquet(
        self,
        table_name: str,
        source_files: list[Path],
        output_dir: Path,
    ) -> Path:
        """Read TBL files, sort, and write as Parquet.

        Args:
            table_name: Benchmark table name (e.g., 'lineitem').
            source_files: One or more TBL source files (may be sharded chunks).
            output_dir: Directory to write the output Parquet file.

        Returns:
            Path to the written Parquet file.

        Raises:
            FileNotFoundError: If any source file does not exist.
            ValueError: If sort columns reference non-existent columns.
        """
        for f in source_files:
            if not f.exists():
                raise FileNotFoundError(f"Source file not found: {f}")

        sort_columns = self.config.get_sort_columns_for_table(table_name)
        column_names = self._get_column_names(table_name)

        # Read all source TBL files into a single table
        arrow_table = self._read_tbl_files(source_files, column_names)

        # Sort if columns are specified
        if sort_columns:
            arrow_table = self._sort_table(arrow_table, sort_columns)

        # Write Parquet
        output_path = output_dir / f"{table_name}.parquet"
        output_dir.mkdir(parents=True, exist_ok=True)

        compression_map = {
            "snappy": "SNAPPY",
            "gzip": "GZIP",
            "zstd": "ZSTD",
            "none": None,
        }
        compression = compression_map.get(self.config.compression)

        pq.write_table(
            arrow_table,
            output_path,
            compression=compression,
            use_dictionary=True,
            write_statistics=True,
            row_group_size=self.config.row_group_size,
        )

        row_count = arrow_table.num_rows
        file_size = output_path.stat().st_size
        logger.info(
            "Wrote sorted Parquet: %s (%s rows, %.1f MB, row_group_size=%d)",
            output_path.name,
            f"{row_count:,}",
            file_size / (1024 * 1024),
            self.config.row_group_size,
        )

        return output_path

    def _get_column_names(self, table_name: str) -> list[str] | None:
        """Get column names for a table from the schema registry.

        Returns:
            List of column names, or None if schema not available.
        """
        schema = self.schema_registry.get(table_name)
        if schema and "columns" in schema:
            return [col["name"] for col in schema["columns"]]
        return None

    def _read_tbl_files(
        self,
        source_files: list[Path],
        column_names: list[str] | None,
    ) -> pa.Table:
        """Read one or more TBL files into a PyArrow table.

        TBL format: pipe-delimited, no header, trailing delimiter on each line.
        """
        read_options = csv.ReadOptions(
            autogenerate_column_names=column_names is None,
            column_names=column_names,
        )
        parse_options = csv.ParseOptions(delimiter="|")

        tables = []
        for source_file in source_files:
            table = csv.read_csv(source_file, read_options=read_options, parse_options=parse_options)
            tables.append(table)

        combined = pa.concat_tables(tables) if len(tables) > 1 else tables[0]

        # TBL files have a trailing pipe which creates an extra empty column.
        # Drop it if present.
        last_col = combined.column_names[-1]
        if last_col.startswith("col_") or last_col == "" or last_col.startswith("f"):
            # Check if the last column is all null/empty (trailing delimiter artifact)
            last_values = combined.column(last_col)
            if last_values.null_count == len(last_values) or (
                pa.types.is_string(last_values.type) and pc.all(pc.equal(last_values, "")).as_py()
            ):
                combined = combined.drop(last_col)

        return combined

    def _sort_table(self, table: pa.Table, sort_columns: list[SortColumn]) -> pa.Table:
        """Sort a PyArrow table by the specified columns.

        Args:
            table: Table to sort.
            sort_columns: Columns and directions to sort by.

        Returns:
            Sorted table.

        Raises:
            ValueError: If a sort column is not found in the table.
        """
        available = set(table.column_names)
        for sc in sort_columns:
            if sc.name not in available:
                raise ValueError(f"Sort column '{sc.name}' not found in table. Available columns: {sorted(available)}")

        sort_keys = [(sc.name, "ascending" if sc.order == SortOrder.ASC else "descending") for sc in sort_columns]
        indices = pc.sort_indices(table, sort_keys=sort_keys)
        return table.take(indices)



"""Compatibility exports for ``benchbox.core.config`` imports.

The canonical models live in :mod:`benchbox.core.schemas`. This module
re-exports them so that existing examples and external code that import
from ``benchbox.core.config`` continue to work.

Copyright 2026 Joe Harris / BenchBox Project

Licensed under the MIT License. See LICENSE file in the project root for details.
"""

from benchbox.core.schemas import BenchmarkConfig, DatabaseConfig, RunConfig

__all__ = ["BenchmarkConfig", "DatabaseConfig", "RunConfig"]



"""Base platform adapter package."""

from __future__ import annotations

from .adapter import (
    BenchmarkResults,
    DriverIsolationCapability,
    EnhancedBenchmarkResults,
    ExecutionPhases,
    PlatformAdapter,
    PlatformAdapterConnection,
    PlatformAdapterCursor,
    QueryDefinition,
    check_isolation_capability,
)
from .models import (
    ConnectionConfig,
    DatabaseValidationResult,
    DataGenerationPhase,
    MaintenanceOperation,
    MaintenanceTestPhase,
    PowerTestPhase,
    QueryExecution,
    SchemaCreationPhase,
    SetupPhase,
    TableCreationStats,
    TableGenerationStats,
    TableLoadingStats,
    ThroughputStream,
    ThroughputTestPhase,
    ValidationPhase,
)
from .utils import FileFormatInfo, detect_file_format, is_non_interactive

__all__ = [
    "ConnectionConfig",
    "DatabaseValidationResult",
    "DataGenerationPhase",
    "BenchmarkResults",
    "DriverIsolationCapability",
    "EnhancedBenchmarkResults",
    "check_isolation_capability",
    "ExecutionPhases",
    "FileFormatInfo",
    "MaintenanceOperation",
    "MaintenanceTestPhase",
    "PlatformAdapter",
    "PlatformAdapterConnection",
    "PlatformAdapterCursor",
    "PowerTestPhase",
    "QueryExecution",
    "SchemaCreationPhase",
    "SetupPhase",
    "TableCreationStats",
    "TableGenerationStats",
    "TableLoadingStats",
    "ThroughputStream",
    "ThroughputTestPhase",
    "ValidationPhase",
    "QueryDefinition",
    "detect_file_format",
    "is_non_interactive",
]



"""Format preference resolution for multi-format manifests."""

from __future__ import annotations

from benchbox.core.manifest.models import ManifestV2
from benchbox.platforms.base.format_capabilities import get_supported_formats


def get_preferred_format(
    manifest: ManifestV2,
    table_name: str,
    platform_name: str,
) -> str | None:
    """Get preferred format for a table on a platform.

    Resolution order:
    1. Manifest format_preference (user-specified)
    2. Platform default preferences
    3. First available format

    Args:
        manifest: ManifestV2 instance
        table_name: Name of the table
        platform_name: Name of the platform (e.g., 'duckdb', 'datafusion')

    Returns:
        Preferred format name if available, None if no formats found
    """
    table_formats = manifest.tables.get(table_name)
    if not table_formats:
        return None

    available_formats = list(table_formats.formats.keys())
    if not available_formats:
        return None

    # Get platform supported formats (in preference order)
    platform_formats = get_supported_formats(platform_name)

    # Try manifest preference first
    for fmt in manifest.format_preference:
        if fmt in available_formats and fmt in platform_formats:
            return fmt

    # Try platform preference
    for fmt in platform_formats:
        if fmt in available_formats:
            return fmt

    # Fallback: first available format
    return available_formats[0]


def get_files_for_format(
    manifest: ManifestV2,
    table_name: str,
    format_name: str,
) -> list[str]:
    """Get file paths for a specific format.

    Args:
        manifest: ManifestV2 instance
        table_name: Name of the table
        format_name: Format to retrieve (e.g., 'parquet', 'delta')

    Returns:
        List of file paths for the specified format, empty list if not found
    """
    table_formats = manifest.tables.get(table_name)
    if not table_formats:
        return []

    files = table_formats.formats.get(format_name, [])
    return [f.path for f in files]


def list_available_formats(manifest: ManifestV2, table_name: str) -> list[str]:
    """List all available formats for a table.

    Args:
        manifest: ManifestV2 instance
        table_name: Name of the table

    Returns:
        List of format names available for the table
    """
    table_formats = manifest.tables.get(table_name)
    if not table_formats:
        return []

    return list(table_formats.formats.keys())



"""SQL-based data generation for TPC-DI."""

from __future__ import annotations

import logging
from typing import Any

from benchbox.utils.clock import elapsed_seconds, mono_time

try:
    import duckdb
except ImportError:  # pragma: no cover - optional dependency
    duckdb = None  # type: ignore[assignment]

logger = logging.getLogger(__name__)


class TPCDISQLGenerator:
    """High-performance TPC-DI data generator using DuckDB SQL for vectorized operations."""

    def __init__(
        self,
        scale_factor: float = 1.0,
        connection: Any | None = None,
        enable_progress: bool = True,
    ):
        """Initialize the SQL-based TPC-DI data generator.

        Args:
            scale_factor: Scale factor for data generation (1.0 = standard size)
            connection: DuckDB connection to use (creates temporary one if None)
            enable_progress: Enable progress logging
        """
        if duckdb is None:
            raise ImportError("DuckDB is required for SQL-based generation")

        self.scale_factor = scale_factor
        self.connection = connection
        self.temp_connection = None
        self.enable_progress = enable_progress

        # Initialize logger
        self.logger = logging.getLogger(self.__class__.__name__)

        # Base sizes for scale_factor = 1.0
        self.base_customers = 50000
        self.base_companies = 1000
        self.base_securities = 10000
        self.base_accounts = 100000
        self.base_trades = 1000000

        # Generation statistics
        self.generation_stats: dict[str, Any] = {
            "records_generated": 0,
            "tables_generated": 0,
            "generation_times": {},
            "sql_operations": 0,
        }

    def _get_connection(self) -> Any:
        """Get DuckDB connection, creating temporary one if needed."""
        if self.connection is not None:
            return self.connection

        if self.temp_connection is None:
            self.temp_connection = duckdb.connect(":memory:")
        return self.temp_connection

    def _close_temp_connection(self):
        """Close temporary connection if created."""
        if self.temp_connection is not None:
            self.temp_connection.close()
            self.temp_connection = None

    def _generate_table(
        self,
        conn: Any,
        table_name: str,
        insert_sql: str,
        num_records: int | None = None,
    ) -> int:
        """Shared scaffolding for dimension/fact table generation.

        Args:
            conn: DuckDB connection.
            table_name: Target table name (used for logging and stats).
            insert_sql: Complete INSERT SQL statement to execute.
            num_records: Pre-computed record count. When None, the count is
                extracted via ``SELECT COUNT(*) FROM <table_name>`` after insert.

        Returns:
            Number of records generated.
        """
        if self.enable_progress:
            if num_records is not None:
                self.logger.info(f"Generating {num_records:,} {table_name} records...")
            else:
                self.logger.info(f"Generating {table_name}...")

        start_time = mono_time()
        conn.execute(insert_sql)

        if num_records is None:
            result = conn.execute(f"SELECT COUNT(*) FROM {table_name}").fetchone()
            num_records = result[0] if result else 0

        generation_time = elapsed_seconds(start_time)
        self.generation_stats["records_generated"] += num_records
        self.generation_stats["tables_generated"] += 1
        self.generation_stats["generation_times"][table_name] = generation_time
        self.generation_stats["sql_operations"] += 1

        if self.enable_progress:
            self.logger.info(f"✅ Generated {num_records:,} {table_name} records in {generation_time:.3f}s")

        return num_records

    def _setup_lookup_tables(self, conn: Any) -> None:
        """Create lookup tables for random data selection."""
        if self.enable_progress:
            self.logger.info("Setting up lookup tables for random data generation...")

        # First names lookup
        conn.execute("""
            CREATE TEMP TABLE IF NOT EXISTS lookup_first_names AS
            SELECT unnest([
                'John', 'Jane', 'Michael', 'Sarah', 'David', 'Lisa', 'Robert', 'Maria',
                'James', 'Jennifer', 'William', 'Patricia', 'Richard', 'Linda', 'Joseph',
                'Barbara', 'Thomas', 'Elizabeth', 'Christopher', 'Helen', 'Charles', 'Sandra',
                'Daniel', 'Donna', 'Matthew', 'Carol', 'Anthony', 'Ruth', 'Mark', 'Sharon',
                'Donald', 'Michelle', 'Steven', 'Laura', 'Paul', 'Sarah', 'Andrew', 'Kimberly'
            ]) as name, row_number() OVER () as id
        """)

        # Last names lookup
        conn.execute("""
            CREATE TEMP TABLE IF NOT EXISTS lookup_last_names AS
            SELECT unnest([
                'Smith', 'Johnson', 'Williams', 'Brown', 'Jones', 'Garcia', 'Miller', 'Davis',
                'Rodriguez', 'Martinez', 'Hernandez', 'Lopez', 'Gonzalez', 'Wilson', 'Anderson',
                'Thomas', 'Taylor', 'Moore', 'Jackson', 'Martin', 'Lee', 'Perez', 'Thompson',
                'White', 'Harris', 'Sanchez', 'Clark', 'Ramirez', 'Lewis', 'Robinson', 'Walker'
            ]) as name, row_number() OVER () as id
        """)

        # Industries lookup
        conn.execute("""
            CREATE TEMP TABLE IF NOT EXISTS lookup_industries AS
            SELECT unnest([
                'Technology', 'Healthcare', 'Financial Services', 'Manufacturing',
                'Retail', 'Energy', 'Telecommunications', 'Transportation',
                'Real Estate', 'Media', 'Utilities', 'Consumer Goods',
                'Aerospace', 'Automotive', 'Biotechnology', 'Construction'
            ]) as industry, row_number() OVER () as id
        """)

        # Company names lookup
        conn.execute("""
            CREATE TEMP TABLE IF NOT EXISTS lookup_company_names AS
            SELECT unnest([
                'TechCorp', 'GlobalInc', 'InnovativeSoft', 'DataSystems', 'CloudTech',
                'SecureNet', 'SmartSolutions', 'DigitalWorks', 'InfoTech', 'CyberCorp',
                'NextGen', 'PowerSystems', 'EliteServices', 'PrimeData', 'CoreTech',
                'VisionSoft', 'AlphaWorks', 'BetaSystems', 'GammaTech', 'DeltaCorp'
            ]) as name, row_number() OVER () as id
        """)

        # SP Ratings lookup
        conn.execute("""
            CREATE TEMP TABLE IF NOT EXISTS lookup_sp_ratings AS
            SELECT unnest([
                'AAA', 'AA+', 'AA', 'AA-', 'A+', 'A', 'A-',
                'BBB+', 'BBB', 'BBB-', 'BB+', 'BB', 'BB-'
            ]) as rating, row_number() OVER () as id
        """)

        # Trade types lookup
        conn.execute("""
            CREATE TEMP TABLE IF NOT EXISTS lookup_trade_types AS
            SELECT unnest([
                'Market Buy', 'Market Sell', 'Limit Buy', 'Limit Sell',
                'Stop Buy', 'Stop Sell', 'Stop Limit Buy', 'Stop Limit Sell'
            ]) as type, row_number() OVER () as id
        """)

        # Status values lookup
        conn.execute("""
            CREATE TEMP TABLE IF NOT EXISTS lookup_statuses AS
            SELECT unnest(['Active', 'Inactive', 'Suspended']) as status, row_number() OVER () as id
        """)

        # US states lookup
        conn.execute("""
            CREATE TEMP TABLE IF NOT EXISTS lookup_us_states AS
            SELECT unnest([
                'CA', 'NY', 'TX', 'FL', 'IL', 'PA', 'OH', 'GA', 'NC', 'MI',
                'NJ', 'VA', 'WA', 'AZ', 'MA', 'TN', 'IN', 'MO', 'MD', 'WI'
            ]) as state, row_number() OVER () as id
        """)

        if self.enable_progress:
            self.logger.info("✅ Lookup tables created")

    def generate_date_dimension(self, conn: Any, start_year: int = 2020, end_year: int = 2024) -> int:
        """Generate date dimension using SQL."""
        insert_sql = f"""
            INSERT INTO DimDate
            SELECT
                row_number() OVER (ORDER BY date_val) as SK_DateID,
                date_val as DateValue,
                strftime(date_val, '%Y-%m-%d') as DateDesc,
                EXTRACT(year FROM date_val) as CalendarYearID,
                CAST(EXTRACT(year FROM date_val) AS VARCHAR) as CalendarYearDesc,
                EXTRACT(quarter FROM date_val) as CalendarQtrID,
                'Q' || EXTRACT(quarter FROM date_val) || ' ' || EXTRACT(year FROM date_val) as CalendarQtrDesc,
                EXTRACT(month FROM date_val) as CalendarMonthID,
                strftime(date_val, '%B %Y') as CalendarMonthDesc,
                EXTRACT(week FROM date_val) as CalendarWeekID,
                'Week ' || EXTRACT(week FROM date_val) || ' ' || EXTRACT(year FROM date_val) as CalendarWeekDesc,
                EXTRACT(dow FROM date_val) + 1 as DayOfWeekNum,
                strftime(date_val, '%A') as DayOfWeekDesc,
                EXTRACT(year FROM date_val) as FiscalYearID,  -- Simplified: same as calendar
                'FY ' || EXTRACT(year FROM date_val) as FiscalYearDesc,
                EXTRACT(quarter FROM date_val) as FiscalQtrID,
                'FY ' || EXTRACT(year FROM date_val) || ' Q' || EXTRACT(quarter FROM date_val) as FiscalQtrDesc,
                CASE WHEN EXTRACT(dow FROM date_val) IN (0, 6) THEN true ELSE false END as HolidayFlag
            FROM (
                SELECT (DATE '{start_year}-01-01' + INTERVAL (s.generate_series) DAY) as date_val
                FROM generate_series(0, {(end_year - start_year + 1) * 365 + 10}) s  -- Add extra days for leap years
            ) dates
            WHERE date_val <= DATE '{end_year}-12-31'
        """
        return self._generate_table(conn, "DimDate", insert_sql)

    def generate_time_dimension(self, conn: Any) -> int:
        """Generate time dimension using SQL."""
        insert_sql = """
            INSERT INTO DimTime
            SELECT
                row_number() OVER () as SK_TimeID,
                make_time(hour_val, minute_val, 0) as TimeValue,
                hour_val as HourID,
                'Hour ' || lpad(CAST(hour_val AS VARCHAR), 2, '0') as HourDesc,
                minute_val as MinuteID,
                'Minute ' || lpad(CAST(minute_val AS VARCHAR), 2, '0') as MinuteDesc,
                0 as SecondID,
                'Second 00' as SecondDesc,
                CASE
                    WHEN (hour_val = 9 AND minute_val >= 30) OR (hour_val > 9 AND hour_val < 16)
                    THEN true ELSE false
                END as MarketHoursFlag,
                CASE WHEN hour_val >= 8 AND hour_val < 18 THEN true ELSE false END as OfficeHoursFlag
            FROM (
                SELECT
                    h.hour_val,
                    m.minute_val
                FROM
                    (SELECT generate_series as hour_val FROM generate_series(0, 23)) h
                CROSS JOIN
                    (SELECT generate_series * 5 as minute_val FROM generate_series(0, 11)) m
            ) times
            ORDER BY hour_val, minute_val
        """
        return self._generate_table(conn, "DimTime", insert_sql)

    def generate_company_dimension(self, conn: Any) -> int:
        """Generate company dimension using SQL."""
        num_companies = int(self.base_companies * self.scale_factor)

        insert_sql = f"""
            INSERT INTO DimCompany
            SELECT
                s.generate_series as SK_CompanyID,
                s.generate_series as CompanyID,
                st.status as Status,
                cn.name || ' ' || s.generate_series as Name,
                ind.industry as Industry,
                rating.rating as SPrating,
                CASE WHEN rating.rating IN ('BBB-', 'BB+', 'BB', 'BB-') THEN true ELSE false END as IsLowGrade,
                'CEO ' || lpad(CAST(s.generate_series AS VARCHAR), 4, '0') as CEO,
                CAST((random() * 9898 + 1)::INTEGER AS VARCHAR) || ' Business Blvd' as AddressLine1,
                CASE WHEN random() < 0.3 THEN 'Suite ' || CAST((random() * 899 + 100)::INTEGER AS VARCHAR) ELSE NULL END as AddressLine2,
                lpad(CAST((random() * 89999 + 10000)::INTEGER AS VARCHAR), 5, '0') as PostalCode,
                'City' || CAST((s.generate_series % 100) AS VARCHAR) as City,
                state.state as StateProv,
                'USA' as Country,
                'Leading company in ' || ind.industry as Description,
                make_date((random() * 70 + 1950)::INTEGER, (random() * 11 + 1)::INTEGER, (random() * 27 + 1)::INTEGER) as FoundingDate,
                true as IsCurrent,
                1 as BatchID,
                DATE '2023-01-01' as EffectiveDate,
                DATE '9999-12-31' as EndDate
            FROM
                generate_series(1, {num_companies}) s(generate_series)
                CROSS JOIN (SELECT status FROM lookup_statuses ORDER BY random() LIMIT 1) st
                CROSS JOIN (SELECT name FROM lookup_company_names ORDER BY random() LIMIT 1) cn
                CROSS JOIN (SELECT industry FROM lookup_industries ORDER BY random() LIMIT 1) ind
                CROSS JOIN (SELECT rating FROM lookup_sp_ratings ORDER BY random() LIMIT 1) rating
                CROSS JOIN (SELECT state FROM lookup_us_states ORDER BY random() LIMIT 1) state
        """
        return self._generate_table(conn, "DimCompany", insert_sql, num_records=num_companies)

    def generate_security_dimension(self, conn: Any) -> int:
        """Generate security dimension using SQL."""
        num_securities = int(self.base_securities * self.scale_factor)

        insert_sql = f"""
            INSERT INTO DimSecurity
            SELECT
                s as SK_SecurityID,
                'SYM' || lpad(CAST(s AS VARCHAR), 4, '0') as Symbol,
                'S' as Issue,  -- Stock
                st.status as Status,
                'Security ' || lpad(CAST(s AS VARCHAR), 4, '0') as Name,
                exchange.exchange_id as ExchangeID,
                (random() * (SELECT MAX(SK_CompanyID) FROM DimCompany) + 1)::INTEGER as SK_CompanyID,
                (random() * (1000000000 - 1000000) + 1000000)::INTEGER as SharesOutstanding,
                make_date((random() * (2020 - 2000) + 2000)::INTEGER, (random() * (12 - 1) + 1)::INTEGER, (random() * (28 - 1) + 1)::INTEGER) as FirstTrade,
                make_date((random() * (2020 - 2000) + 2000)::INTEGER, (random() * (12 - 1) + 1)::INTEGER, (random() * (28 - 1) + 1)::INTEGER) as FirstTradeOnExchange,
                round((random() * 5.0), 2) as Dividend,
                true as IsCurrent,
                1 as BatchID,
                DATE '2023-01-01' as EffectiveDate,
                DATE '9999-12-31' as EndDate
            FROM
                generate_series(1, {num_securities}) s
                CROSS JOIN (SELECT status FROM lookup_statuses ORDER BY random() LIMIT 1) st
                CROSS JOIN (SELECT unnest(['NYSE', 'NASDAQ', 'AMEX']) as exchange_id ORDER BY random() LIMIT 1) exchange
        """
        return self._generate_table(conn, "DimSecurity", insert_sql, num_records=num_securities)

    def generate_customer_dimension(self, conn: Any) -> int:
        """Generate customer dimension using SQL."""
        num_customers = int(self.base_customers * self.scale_factor)

        insert_sql = f"""
            INSERT INTO DimCustomer
            SELECT
                s as SK_CustomerID,
                s as CustomerID,
                'TAX' || lpad(CAST(s AS VARCHAR), 6, '0') as TaxID,
                st.status as Status,
                ln.name as LastName,
                fn.name as FirstName,
                chr((random() * (90 - 65) + 65)::INTEGER) as MiddleInitial,  -- Random A-Z
                CASE WHEN random() < 0.5 THEN 'M' ELSE 'F' END as Gender,
                (random() * (3 - 1) + 1)::INTEGER as Tier,
                make_date((random() * (2000 - 1950) + 1950)::INTEGER, (random() * (12 - 1) + 1)::INTEGER, (random() * (28 - 1) + 1)::INTEGER) as DOB,
                CAST((random() * (9999 - 100) + 100)::INTEGER AS INTEGER) || ' Customer St' as AddressLine1,
                CASE WHEN random() < 0.3 THEN 'Apt ' || CAST((random() * (999 - 1) + 1)::INTEGER AS INTEGER) ELSE NULL END as AddressLine2,
                lpad(CAST((random() * (99999 - 10000) + 10000)::INTEGER AS INTEGER), 5, '0') as PostalCode,
                'City' || CAST((s::INTEGER % 500) AS VARCHAR) as City,
                state.state as StateProv,
                'USA' as Country,
                '555-' || lpad(CAST((random() * (999 - 100) + 100)::INTEGER AS INTEGER), 3, '0') || '-' ||
                lpad(CAST((random() * (9999 - 1000) + 1000)::INTEGER AS INTEGER), 4, '0') as Phone1,
                CASE WHEN random() < 0.2 THEN
                    '555-' || lpad(CAST((random() * (999 - 100) + 100)::INTEGER AS INTEGER), 3, '0') || '-' ||
                    lpad(CAST((random() * (9999 - 1000) + 1000)::INTEGER AS INTEGER), 4, '0')
                ELSE NULL END as Phone2,
                NULL as Phone3,
                'customer' || s || '@example.com' as Email1,
                CASE WHEN random() < 0.1 THEN 'customer' || s || '.alt@example.com' ELSE NULL END as Email2,
                'Federal Tax' as NationalTaxRateDesc,
                0.25 as NationalTaxRate,
                'State Tax' as LocalTaxRateDesc,
                0.08 as LocalTaxRate,
                'AGENCY' || lpad(CAST((random() * (10 - 1) + 1)::INTEGER AS INTEGER), 2, '0') as AgencyID,
                (random() * (850 - 300) + 300)::INTEGER as CreditRating,
                (random() * (10000000 - 10000) + 10000)::INTEGER as NetWorth,
                'Customer Segment ' || (random() * (5 - 1) + 1)::INTEGER as MarketingNameplate,
                true as IsCurrent,
                1 as BatchID,
                DATE '2023-01-01' as EffectiveDate,
                DATE '9999-12-31' as EndDate
            FROM
                generate_series(1, {num_customers}) s
                CROSS JOIN (SELECT status FROM lookup_statuses ORDER BY random() LIMIT 1) st
                CROSS JOIN (SELECT name FROM lookup_first_names ORDER BY random() LIMIT 1) fn
                CROSS JOIN (SELECT name FROM lookup_last_names ORDER BY random() LIMIT 1) ln
                CROSS JOIN (SELECT state FROM lookup_us_states ORDER BY random() LIMIT 1) state
        """
        return self._generate_table(conn, "DimCustomer", insert_sql, num_records=num_customers)

    def generate_account_dimension(self, conn: Any) -> int:
        """Generate account dimension using SQL."""
        num_customers = int(self.base_customers * self.scale_factor)
        num_accounts = int(num_customers * 1.5)  # 1.5 accounts per customer average

        insert_sql = f"""
            INSERT INTO DimAccount
            SELECT
                s as SK_AccountID,
                s as AccountID,
                (random() * (100 - 1) + 1)::INTEGER as SK_BrokerID,  -- Assume 100 brokers
                (random() * (SELECT MAX(SK_CustomerID) FROM DimCustomer) + 1)::INTEGER as SK_CustomerID,
                st.status as Status,
                'Account ' || lpad(CAST(s AS VARCHAR), 6, '0') as AccountDesc,
                (random() * (2 - 0) + 0)::INTEGER as TaxStatus,  -- 0=Taxable, 1=Tax Deferred, 2=Tax Free
                true as IsCurrent,
                1 as BatchID,
                DATE '2023-01-01' as EffectiveDate,
                DATE '9999-12-31' as EndDate
            FROM
                generate_series(1, {num_accounts}) s
                CROSS JOIN (SELECT status FROM lookup_statuses ORDER BY random() LIMIT 1) st
        """
        return self._generate_table(conn, "DimAccount", insert_sql, num_records=num_accounts)

    def generate_trade_facts(self, conn: Any) -> int:
        """Generate trade fact table using SQL."""
        num_trades = int(self.base_trades * self.scale_factor)

        insert_sql = f"""
            INSERT INTO FactTrade
            SELECT
                s as TradeID,
                (random() * (100 - 1) + 1)::INTEGER as SK_BrokerID,
                (random() * (SELECT MAX(SK_DateID) FROM DimDate) + 1)::INTEGER as SK_CreateDateID,
                (random() * (SELECT MAX(SK_TimeID) FROM DimTime) + 1)::INTEGER as SK_CreateTimeID,
                CASE WHEN random() < 0.8 THEN
                    (random() * (SELECT MAX(SK_DateID) FROM DimDate) + 1)::INTEGER
                ELSE NULL END as SK_CloseDateID,
                CASE WHEN random() < 0.8 THEN
                    (random() * (SELECT MAX(SK_TimeID) FROM DimTime) + 1)::INTEGER
                ELSE NULL END as SK_CloseTimeID,
                CASE
                    WHEN random() < 0.8 THEN 'Completed'
                    WHEN random() < 0.9 THEN 'Pending'
                    ELSE 'Cancelled'
                END as Status,
                tt.type as Type,
                CASE WHEN random() < 0.5 THEN true ELSE false END as CashFlag,
                (random() * (SELECT MAX(SK_SecurityID) FROM DimSecurity) + 1)::INTEGER as SK_SecurityID,
                (random() * (SELECT MAX(SK_CompanyID) FROM DimCompany) + 1)::INTEGER as SK_CompanyID,
                (random() * (10000 - 1) + 1)::INTEGER as Quantity,
                round((random() * 490.0 + 10.0), 2) as BidPrice,
                (random() * (SELECT MAX(SK_CustomerID) FROM DimCustomer) + 1)::INTEGER as SK_CustomerID,
                (random() * (SELECT MAX(SK_AccountID) FROM DimAccount) + 1)::INTEGER as SK_AccountID,
                'Broker' || (random() * (100 - 1) + 1)::INTEGER as ExecutedBy,
                round((random() * 490.0 + 10.0) * (random() * 0.04 + 0.98), 2) as TradePrice,
                round((random() * 45.0 + 5.0), 2) as Fee,
                round((random() * 490.0 + 10.0) * (random() * (10000 - 1) + 1)::INTEGER * 0.001, 2) as Commission,
                CASE WHEN random() < 0.8 THEN
                    round((random() * 490.0 + 10.0) * (random() * (10000 - 1) + 1)::INTEGER * 0.01, 2)
                ELSE 0.0 END as Tax,
                1 as BatchID
            FROM
                generate_series(1, {num_trades}) s
                CROSS JOIN (SELECT type FROM lookup_trade_types ORDER BY random() LIMIT 1) tt
        """
        return self._generate_table(conn, "FactTrade", insert_sql, num_records=num_trades)

    def generate_all_tables(self, conn: Any, tables: list[str] | None = None) -> dict[str, int]:
        """Generate all TPC-DI tables using SQL."""
        if tables is None:
            tables = [
                "DimDate",
                "DimTime",
                "DimCompany",
                "DimSecurity",
                "DimCustomer",
                "DimAccount",
                "FactTrade",
            ]

        if self.enable_progress:
            self.logger.info(f"Starting SQL-based TPC-DI data generation (Scale Factor: {self.scale_factor})")

        start_time = mono_time()

        # Setup lookup tables first
        self._setup_lookup_tables(conn)

        results = {}

        # Generate tables in dependency order
        table_generators = {
            "DimDate": self.generate_date_dimension,
            "DimTime": self.generate_time_dimension,
            "DimCompany": self.generate_company_dimension,
            "DimSecurity": self.generate_security_dimension,
            "DimCustomer": self.generate_customer_dimension,
            "DimAccount": self.generate_account_dimension,
            "FactTrade": self.generate_trade_facts,
        }

        for table_name in tables:
            if table_name in table_generators:
                table_start = mono_time()
                records_generated = table_generators[table_name](conn)
                results[table_name] = records_generated

                table_time = elapsed_seconds(table_start)
                if self.enable_progress:
                    throughput = records_generated / table_time if table_time > 0 else 0
                    self.logger.info(f"  {table_name}: {throughput:,.0f} records/second")

        total_time = elapsed_seconds(start_time)
        total_records = sum(results.values())

        if self.enable_progress:
            overall_throughput = total_records / total_time if total_time > 0 else 0
            self.logger.info(
                f"SQL generation completed: {total_records:,} records in {total_time:.3f}s ({overall_throughput:,.0f} records/sec)"
            )

        return results

    def get_generation_stats(self) -> dict[str, Any]:
        """Get generation statistics."""
        return {
            **self.generation_stats,
            "scale_factor": self.scale_factor,
            "estimated_records": {
                "DimCustomer": int(self.base_customers * self.scale_factor),
                "DimAccount": int(self.base_accounts * self.scale_factor),
                "DimSecurity": int(self.base_securities * self.scale_factor),
                "DimCompany": int(self.base_companies * self.scale_factor),
                "FactTrade": int(self.base_trades * self.scale_factor),
            },
        }

    def __del__(self):
        """Cleanup temporary connection on destruction."""
        self._close_temp_connection()


__all__ = ["TPCDISQLGenerator"]



"""Shared catalog loading functions for primitives benchmarks.

Provides generic YAML catalog loaders that handle the two catalog patterns:
- Operations catalogs (write_primitives, transaction_primitives): entries with write_sql,
  validation_queries, cleanup_sql, file_dependencies, platform_overrides, requires_setup
- Query catalogs (read_primitives, metadata_primitives): entries with sql, variants, skip_on

Each primitives module defines its own dataclasses and error types, then delegates
the actual YAML parsing and validation to these shared functions.

Copyright 2026 Joe Harris / BenchBox Project

Licensed under the MIT License. See LICENSE file in the project root for details.
"""

from __future__ import annotations

from importlib import resources
from typing import Any

import yaml


def load_operations_catalog(
    *,
    package: str,
    catalog_filename: str,
    error_class: type[RuntimeError],
    label: str,
    build_validation_query: Any,
    build_operation: Any,
    build_catalog: Any,
) -> Any:
    """Load and validate an operations-style catalog from package resources.

    This handles the common loading pattern shared by write_primitives and
    transaction_primitives catalogs.

    Args:
        package: The ``__package__`` of the calling module (for resource lookup)
        catalog_filename: YAML filename to load (e.g., "operations.yaml")
        error_class: Exception class to raise on validation failures
        label: Human-readable label for error messages (e.g., "Write Primitives")
        build_validation_query: Callable that builds a ValidationQuery from parsed fields
        build_operation: Callable that builds a WriteOperation from parsed fields
        build_catalog: Callable(version, operations) that builds the catalog container

    Returns:
        The catalog container built by ``build_catalog``

    Raises:
        error_class: If catalog cannot be loaded or is invalid
    """
    payload = _load_yaml(package, catalog_filename, error_class, label, "operation")

    version = _parse_version(payload, error_class, label, "operation")

    raw_entries = payload.get("operations")
    if not isinstance(raw_entries, list):
        raise error_class(f"{label} operation catalog must define an 'operations' list")

    operations: dict[str, Any] = {}

    for index, entry in enumerate(raw_entries):
        if not isinstance(entry, dict):
            raise error_class(f"Catalog entry at index {index} must be a mapping")

        operation_id = _parse_required_string(entry, "id", index, None, error_class)

        if operation_id in operations:
            raise error_class(f"Duplicate operation id detected in catalog: {operation_id}")

        operations[operation_id] = _parse_operation_entry(
            entry, index, operation_id, error_class, build_validation_query, build_operation
        )

    return build_catalog(version=version, operations=operations)


def _parse_operation_entry(
    entry: dict[str, Any],
    index: int,
    operation_id: str,
    error_class: type[RuntimeError],
    build_validation_query: Any,
    build_operation: Any,
) -> Any:
    """Parse and validate a single operation entry from the catalog YAML."""
    category = entry.get("category")
    if not isinstance(category, str) or not category.strip():
        category = operation_id.split("_")[0]
    category = category.strip().lower()

    description = entry.get("description")
    if not isinstance(description, str) or not description.strip():
        raise error_class(f"Catalog entry '{operation_id}' must include a description")
    description = description.strip()

    write_sql = entry.get("write_sql")
    if not isinstance(write_sql, str) or not write_sql.strip():
        raise error_class(f"Catalog entry '{operation_id}' must include non-empty write_sql")

    validation_queries = _parse_validation_queries(entry, operation_id, error_class, build_validation_query)

    cleanup_sql = entry.get("cleanup_sql")
    if cleanup_sql is not None and not isinstance(cleanup_sql, str):
        raise error_class(f"Catalog entry '{operation_id}' cleanup_sql must be a string")

    expected_rows_affected = entry.get("expected_rows_affected")
    if expected_rows_affected is not None:
        try:
            expected_rows_affected = int(expected_rows_affected)
        except (TypeError, ValueError) as exc:
            raise error_class(f"Catalog entry '{operation_id}' expected_rows_affected must be an integer") from exc

    file_dependencies = entry.get("file_dependencies", [])
    if not isinstance(file_dependencies, list):
        raise error_class(f"Catalog entry '{operation_id}' file_dependencies must be a list")

    platform_overrides = entry.get("platform_overrides", {})
    if not isinstance(platform_overrides, dict):
        raise error_class(f"Catalog entry '{operation_id}' platform_overrides must be a mapping")

    requires_setup = entry.get("requires_setup", True)
    if not isinstance(requires_setup, bool):
        raise error_class(f"Catalog entry '{operation_id}' requires_setup must be a boolean")

    return build_operation(
        id=operation_id,
        category=category,
        description=description,
        write_sql=write_sql,
        validation_queries=validation_queries,
        cleanup_sql=cleanup_sql,
        expected_rows_affected=expected_rows_affected,
        file_dependencies=file_dependencies,
        platform_overrides=platform_overrides,
        requires_setup=requires_setup,
    )


def load_query_catalog(
    *,
    package: str,
    catalog_filename: str,
    error_class: type[RuntimeError],
    label: str,
    build_query: Any,
    build_catalog: Any,
) -> Any:
    """Load and validate a query-style catalog from package resources.

    This handles the common loading pattern shared by read_primitives and
    metadata_primitives catalogs.

    Args:
        package: The ``__package__`` of the calling module (for resource lookup)
        catalog_filename: YAML filename to load (e.g., "queries.yaml")
        error_class: Exception class to raise on validation failures
        label: Human-readable label for error messages (e.g., "Metadata Primitives")
        build_query: Callable that builds a query dataclass from parsed fields
        build_catalog: Callable(version, queries) that builds the catalog container

    Returns:
        The catalog container built by ``build_catalog``

    Raises:
        error_class: If catalog cannot be loaded or is invalid
    """
    payload = _load_yaml(package, catalog_filename, error_class, label, "query")

    version = _parse_version(payload, error_class, label, "query")

    raw_entries = payload.get("queries")
    if not isinstance(raw_entries, list):
        raise error_class(f"{label} query catalog must define a 'queries' list")

    queries: dict[str, Any] = {}

    for index, entry in enumerate(raw_entries):
        if not isinstance(entry, dict):
            raise error_class(f"Catalog entry at index {index} must be a mapping")

        query_id = _parse_required_string(entry, "id", index, None, error_class)

        if query_id in queries:
            raise error_class(f"Duplicate query id detected in catalog: {query_id}")

        raw_sql = entry.get("sql")
        if not isinstance(raw_sql, str) or not raw_sql.strip():
            raise error_class(f"Catalog entry '{query_id}' must include non-empty SQL text")

        category = entry.get("category")
        if category is None:
            category = query_id.split("_", 1)[0]
        if not isinstance(category, str) or not category.strip():
            raise error_class(f"Catalog entry '{query_id}' must define a non-empty category")
        category = category.strip().lower()

        description = entry.get("description")
        if description is not None:
            if not isinstance(description, str) or not description.strip():
                raise error_class(
                    f"Catalog entry '{query_id}' has invalid 'description'; expected non-empty string",
                )
            description = description.strip()

        variants = _parse_variants(entry, query_id, error_class)
        skip_on = _parse_skip_on(entry, query_id, error_class)

        queries[query_id] = build_query(
            id=query_id,
            category=category,
            sql=raw_sql,
            description=description,
            variants=variants,
            skip_on=skip_on,
        )

    return build_catalog(version=version, queries=queries)


# ---------------------------------------------------------------------------
# Internal helpers
# ---------------------------------------------------------------------------


def _load_yaml(
    package: str,
    filename: str,
    error_class: type[RuntimeError],
    label: str,
    kind: str,
) -> dict[str, Any]:
    """Load and parse a YAML catalog file from package resources."""
    try:
        catalog_file = resources.files(package).joinpath(filename)
    except (AttributeError, FileNotFoundError) as exc:
        raise error_class(f"{label} {kind} catalog resource not found") from exc

    try:
        with catalog_file.open("r", encoding="utf-8") as handle:
            payload = yaml.safe_load(handle) or {}
    except OSError as exc:
        raise error_class(f"Unable to read {label.lower()} {kind} catalog") from exc
    except yaml.YAMLError as exc:
        raise error_class(f"Invalid YAML in {label.lower()} {kind} catalog") from exc

    if not isinstance(payload, dict):
        raise error_class(f"{label} {kind} catalog must be a mapping")

    return payload


def _parse_version(
    payload: dict[str, Any],
    error_class: type[RuntimeError],
    label: str,
    kind: str,
) -> int:
    """Parse and validate the catalog version field."""
    raw_version = payload.get("version", 1)
    try:
        return int(raw_version)
    except (TypeError, ValueError) as exc:
        raise error_class(f"{label} {kind} catalog version must be an integer") from exc


def _parse_required_string(
    entry: dict[str, Any],
    field: str,
    index: int,
    entry_id: str | None,
    error_class: type[RuntimeError],
) -> str:
    """Parse a required non-empty string field from a catalog entry."""
    value = entry.get(field)
    if not isinstance(value, str) or not value.strip():
        context = f"'{entry_id}'" if entry_id else f"at index {index}"
        raise error_class(f"Catalog entry {context} is missing a valid '{field}'")
    return value.strip()


def _parse_validation_queries(
    entry: dict[str, Any],
    operation_id: str,
    error_class: type[RuntimeError],
    build_validation_query: Any,
) -> list[Any]:
    """Parse validation queries from an operation entry."""
    raw_validations = entry.get("validation_queries", [])
    if not isinstance(raw_validations, list):
        raise error_class(f"Catalog entry '{operation_id}' validation_queries must be a list")

    validation_queries = []
    for val_idx, val_entry in enumerate(raw_validations):
        if not isinstance(val_entry, dict):
            raise error_class(f"Validation query {val_idx} in operation '{operation_id}' must be a mapping")

        val_id = val_entry.get("id")
        if not isinstance(val_id, str) or not val_id.strip():
            raise error_class(f"Validation query {val_idx} in operation '{operation_id}' missing 'id'")

        val_sql = val_entry.get("sql")
        if not isinstance(val_sql, str) or not val_sql.strip():
            raise error_class(f"Validation query '{val_id}' in operation '{operation_id}' missing 'sql'")

        validation_queries.append(
            build_validation_query(
                id=val_id.strip(),
                sql=val_sql,
                expected_rows=val_entry.get("expected_rows"),
                expected_rows_min=val_entry.get("expected_rows_min"),
                expected_rows_max=val_entry.get("expected_rows_max"),
                expected_values=val_entry.get("expected_values"),
                check_expression=val_entry.get("check_expression"),
            )
        )

    return validation_queries


def _parse_variants(
    entry: dict[str, Any],
    query_id: str,
    error_class: type[RuntimeError],
) -> dict[str, str] | None:
    """Parse dialect-specific variants from a query entry."""
    raw_variants = entry.get("variants")
    if raw_variants is None:
        return None

    if not isinstance(raw_variants, dict):
        raise error_class(f"Catalog entry '{query_id}' has invalid 'variants'; expected mapping of dialect -> SQL")

    variants = {}
    for dialect, variant_sql in raw_variants.items():
        if not isinstance(dialect, str) or not dialect.strip():
            raise error_class(f"Catalog entry '{query_id}' variant dialect must be a non-empty string")
        if not isinstance(variant_sql, str) or not variant_sql.strip():
            raise error_class(f"Catalog entry '{query_id}' variant SQL for dialect '{dialect}' must be non-empty")
        variants[dialect.lower().strip()] = variant_sql

    return variants


def _parse_skip_on(
    entry: dict[str, Any],
    query_id: str,
    error_class: type[RuntimeError],
) -> list[str] | None:
    """Parse skip_on dialect list from a query entry."""
    raw_skip_on = entry.get("skip_on")
    if raw_skip_on is None:
        return None

    if not isinstance(raw_skip_on, list):
        raise error_class(f"Catalog entry '{query_id}' has invalid 'skip_on'; expected list of dialects")

    skip_on = []
    for dialect in raw_skip_on:
        if not isinstance(dialect, str) or not dialect.strip():
            raise error_class(f"Catalog entry '{query_id}' skip_on dialect must be a non-empty string")
        skip_on.append(dialect.lower().strip())

    return skip_on


__all__ = [
    "load_operations_catalog",
    "load_query_catalog",
]



"""Write Primitives benchmark schema definitions.

This module defines the schema for the Write Primitives benchmark, including:
- Base TPC-H tables (reused from tpch.schema)
- Staging tables for write operations
- Metadata tables for tracking operations

Copyright 2026 Joe Harris / BenchBox Project

This implementation is derived from TPC Benchmark™ H (TPC-H) - Copyright © Transaction Processing Performance Council

Licensed under the MIT License. See LICENSE file in the project root for details.
"""

from typing import Any, cast

# Import base TPC-H tables
from benchbox.core.tpch.schema import (
    CUSTOMER,
    LINEITEM,
    NATION,
    ORDERS,
    PART,
    PARTSUPP,
    REGION,
    SUPPLIER,
)

# Category-based staging tables for Write Primitives v2
# Each category gets dedicated staging tables to avoid shared state and enable
# non-transactional cleanup through explicit SQL statements.

# INSERT operations staging table for lineitem
INSERT_OPS_LINEITEM = {
    "name": "insert_ops_lineitem",
    "columns": [
        {"name": "l_orderkey", "type": "INTEGER"},
        {"name": "l_partkey", "type": "INTEGER"},
        {"name": "l_suppkey", "type": "INTEGER"},
        {"name": "l_linenumber", "type": "INTEGER"},
        {"name": "l_quantity", "type": "DECIMAL(15,2)"},
        {"name": "l_extendedprice", "type": "DECIMAL(15,2)"},
        {"name": "l_discount", "type": "DECIMAL(15,2)"},
        {"name": "l_tax", "type": "DECIMAL(15,2)"},
        {"name": "l_returnflag", "type": "VARCHAR(1)"},
        {"name": "l_linestatus", "type": "VARCHAR(1)"},
        {"name": "l_shipdate", "type": "DATE"},
        {"name": "l_commitdate", "type": "DATE"},
        {"name": "l_receiptdate", "type": "DATE"},
        {"name": "l_shipinstruct", "type": "VARCHAR(25)"},
        {"name": "l_shipmode", "type": "VARCHAR(10)"},
        {"name": "l_comment", "type": "VARCHAR(44)"},
    ],
    "primary_key": ["l_orderkey", "l_linenumber"],
}

# INSERT operations staging table for orders
INSERT_OPS_ORDERS = {
    "name": "insert_ops_orders",
    "columns": [
        {"name": "o_orderkey", "type": "INTEGER", "primary_key": True},
        {"name": "o_custkey", "type": "INTEGER"},
        {"name": "o_orderstatus", "type": "VARCHAR(1)"},
        {"name": "o_totalprice", "type": "DECIMAL(15,2)"},
        {"name": "o_orderdate", "type": "DATE"},
        {"name": "o_orderpriority", "type": "VARCHAR(15)"},
        {"name": "o_clerk", "type": "VARCHAR(15)"},
        {"name": "o_shippriority", "type": "INTEGER"},
        {"name": "o_comment", "type": "VARCHAR(79)"},
    ],
}

# INSERT operations staging table for summary/aggregated data
INSERT_OPS_ORDERS_SUMMARY = {
    "name": "insert_ops_orders_summary",
    "columns": [
        {"name": "o_orderkey", "type": "INTEGER", "nullable": True},
        {"name": "o_custkey", "type": "INTEGER", "nullable": True},
        {"name": "o_orderdate", "type": "DATE", "nullable": True},
        {"name": "o_totalprice", "type": "DECIMAL(15,2)", "nullable": True},
        {"name": "customer_name", "type": "VARCHAR(25)", "nullable": True},
        {"name": "order_count", "type": "INTEGER", "nullable": True},
    ],
}

# UPDATE operations staging table (populated from orders during setup)
UPDATE_OPS_ORDERS = {
    "name": "update_ops_orders",
    "columns": [
        {"name": "o_orderkey", "type": "INTEGER", "primary_key": True},
        {"name": "o_custkey", "type": "INTEGER"},
        {"name": "o_orderstatus", "type": "VARCHAR(1)"},
        {"name": "o_totalprice", "type": "DECIMAL(15,2)"},
        {"name": "o_orderdate", "type": "DATE"},
        {"name": "o_orderpriority", "type": "VARCHAR(15)"},
        {"name": "o_clerk", "type": "VARCHAR(15)"},
        {"name": "o_shippriority", "type": "INTEGER"},
        {"name": "o_comment", "type": "VARCHAR(79)"},
    ],
}

# DELETE operations staging table for orders (populated from orders during setup)
DELETE_OPS_ORDERS = {
    "name": "delete_ops_orders",
    "columns": [
        {"name": "o_orderkey", "type": "INTEGER", "primary_key": True},
        {"name": "o_custkey", "type": "INTEGER"},
        {"name": "o_orderstatus", "type": "VARCHAR(1)"},
        {"name": "o_totalprice", "type": "DECIMAL(15,2)"},
        {"name": "o_orderdate", "type": "DATE"},
        {"name": "o_orderpriority", "type": "VARCHAR(15)"},
        {"name": "o_clerk", "type": "VARCHAR(15)"},
        {"name": "o_shippriority", "type": "INTEGER"},
        {"name": "o_comment", "type": "VARCHAR(79)"},
    ],
}

# DELETE operations staging table for lineitem (populated from lineitem during setup)
DELETE_OPS_LINEITEM = {
    "name": "delete_ops_lineitem",
    "columns": [
        {"name": "l_orderkey", "type": "INTEGER"},
        {"name": "l_partkey", "type": "INTEGER"},
        {"name": "l_suppkey", "type": "INTEGER"},
        {"name": "l_linenumber", "type": "INTEGER"},
        {"name": "l_quantity", "type": "DECIMAL(15,2)"},
        {"name": "l_extendedprice", "type": "DECIMAL(15,2)"},
        {"name": "l_discount", "type": "DECIMAL(15,2)"},
        {"name": "l_tax", "type": "DECIMAL(15,2)"},
        {"name": "l_returnflag", "type": "VARCHAR(1)"},
        {"name": "l_linestatus", "type": "VARCHAR(1)"},
        {"name": "l_shipdate", "type": "DATE"},
        {"name": "l_commitdate", "type": "DATE"},
        {"name": "l_receiptdate", "type": "DATE"},
        {"name": "l_shipinstruct", "type": "VARCHAR(25)"},
        {"name": "l_shipmode", "type": "VARCHAR(10)"},
        {"name": "l_comment", "type": "VARCHAR(44)"},
    ],
    "primary_key": ["l_orderkey", "l_linenumber"],
}

# DELETE operations staging table for supplier (for GDPR operations)
DELETE_OPS_SUPPLIER = {
    "name": "delete_ops_supplier",
    "columns": [
        {"name": "s_suppkey", "type": "INTEGER", "primary_key": True},
        {"name": "s_name", "type": "VARCHAR(25)"},
        {"name": "s_address", "type": "VARCHAR(40)"},
        {"name": "s_nationkey", "type": "INTEGER"},
        {"name": "s_phone", "type": "VARCHAR(15)"},
        {"name": "s_acctbal", "type": "DECIMAL(15,2)"},
        {"name": "s_comment", "type": "VARCHAR(101)"},
    ],
}

# MERGE operations target table
MERGE_OPS_TARGET = {
    "name": "merge_ops_target",
    "columns": [
        {"name": "o_orderkey", "type": "INTEGER", "primary_key": True},
        {"name": "o_custkey", "type": "INTEGER"},
        {"name": "o_orderstatus", "type": "VARCHAR(1)"},
        {"name": "o_totalprice", "type": "DECIMAL(15,2)"},
        {"name": "o_orderdate", "type": "DATE"},
        {"name": "o_orderpriority", "type": "VARCHAR(15)"},
        {"name": "o_clerk", "type": "VARCHAR(15)"},
        {"name": "o_shippriority", "type": "INTEGER"},
        {"name": "o_comment", "type": "VARCHAR(79)"},
    ],
}

# MERGE operations source table
MERGE_OPS_SOURCE = {
    "name": "merge_ops_source",
    "columns": [
        {"name": "o_orderkey", "type": "INTEGER", "primary_key": True},
        {"name": "o_custkey", "type": "INTEGER"},
        {"name": "o_orderstatus", "type": "VARCHAR(1)"},
        {"name": "o_totalprice", "type": "DECIMAL(15,2)"},
        {"name": "o_orderdate", "type": "DATE"},
        {"name": "o_orderpriority", "type": "VARCHAR(15)"},
        {"name": "o_clerk", "type": "VARCHAR(15)"},
        {"name": "o_shippriority", "type": "INTEGER"},
        {"name": "o_comment", "type": "VARCHAR(79)"},
    ],
}

# MERGE operations target table for lineitem
MERGE_OPS_LINEITEM_TARGET = {
    "name": "merge_ops_lineitem_target",
    "columns": [
        {"name": "l_orderkey", "type": "INTEGER"},
        {"name": "l_partkey", "type": "INTEGER"},
        {"name": "l_suppkey", "type": "INTEGER"},
        {"name": "l_linenumber", "type": "INTEGER"},
        {"name": "l_quantity", "type": "DECIMAL(15,2)"},
        {"name": "l_extendedprice", "type": "DECIMAL(15,2)"},
        {"name": "l_discount", "type": "DECIMAL(15,2)"},
        {"name": "l_tax", "type": "DECIMAL(15,2)"},
        {"name": "l_returnflag", "type": "VARCHAR(1)"},
        {"name": "l_linestatus", "type": "VARCHAR(1)"},
        {"name": "l_shipdate", "type": "DATE"},
        {"name": "l_commitdate", "type": "DATE"},
        {"name": "l_receiptdate", "type": "DATE"},
        {"name": "l_shipinstruct", "type": "VARCHAR(25)"},
        {"name": "l_shipmode", "type": "VARCHAR(10)"},
        {"name": "l_comment", "type": "VARCHAR(44)"},
    ],
    "primary_key": ["l_orderkey", "l_linenumber"],
}

# MERGE operations target table for summary data
MERGE_OPS_SUMMARY_TARGET = {
    "name": "merge_ops_summary_target",
    "columns": [
        {"name": "o_orderkey", "type": "INTEGER", "nullable": True},
        {"name": "o_custkey", "type": "INTEGER", "nullable": True},
        {"name": "o_orderdate", "type": "DATE", "nullable": True},
        {"name": "o_totalprice", "type": "DECIMAL(15,2)", "nullable": True},
        {"name": "customer_name", "type": "VARCHAR(25)", "nullable": True},
        {"name": "order_count", "type": "INTEGER", "nullable": True},
    ],
}

# INSERT operations enriched table for joined data
INSERT_OPS_LINEITEM_ENRICHED = {
    "name": "insert_ops_lineitem_enriched",
    "columns": [
        # Core lineitem columns
        {"name": "l_orderkey", "type": "INTEGER"},
        {"name": "l_partkey", "type": "INTEGER"},
        {"name": "l_suppkey", "type": "INTEGER"},
        {"name": "l_quantity", "type": "DECIMAL(15,2)"},
        {"name": "l_extendedprice", "type": "DECIMAL(15,2)"},
        # Enriched columns from joins
        {"name": "o_orderdate", "type": "DATE"},
        {"name": "c_name", "type": "VARCHAR(25)"},
        {"name": "s_name", "type": "VARCHAR(25)"},
        {"name": "p_name", "type": "VARCHAR(55)"},
    ],
}

# BULK_LOAD operations target table
BULK_LOAD_OPS_TARGET = {
    "name": "bulk_load_ops_target",
    "columns": [
        {"name": "o_orderkey", "type": "INTEGER"},
        {"name": "o_custkey", "type": "INTEGER"},
        {"name": "o_orderstatus", "type": "VARCHAR(1)"},
        {"name": "o_totalprice", "type": "DECIMAL(15,2)"},
        {"name": "o_orderdate", "type": "DATE"},
        {"name": "o_orderpriority", "type": "VARCHAR(15)", "nullable": True},
        {"name": "o_clerk", "type": "VARCHAR(15)", "nullable": True},
        {"name": "o_shippriority", "type": "INTEGER", "nullable": True},
        {"name": "o_comment", "type": "VARCHAR(79)", "nullable": True},
    ],
}

# DDL operations target table for TRUNCATE tests
DDL_TRUNCATE_TARGET = {
    "name": "ddl_truncate_target",
    "columns": [
        {"name": "o_orderkey", "type": "INTEGER", "primary_key": True},
        {"name": "o_custkey", "type": "INTEGER"},
        {"name": "o_orderdate", "type": "DATE"},
    ],
}

# Audit log for tracking write operations
WRITE_OPS_LOG = {
    "name": "write_ops_log",
    "columns": [
        {"name": "log_id", "type": "INTEGER", "primary_key": True},
        {"name": "operation_id", "type": "VARCHAR(100)"},
        {"name": "operation_category", "type": "VARCHAR(50)"},
        {"name": "timestamp", "type": "TIMESTAMP"},
        {"name": "rows_affected", "type": "INTEGER"},
        {"name": "duration_ms", "type": "DECIMAL(15,2)"},
        {"name": "success", "type": "BOOLEAN"},
        {"name": "error_message", "type": "VARCHAR(500)", "nullable": True},
    ],
}

# Batch metadata for bulk operations
BATCH_METADATA = {
    "name": "batch_metadata",
    "columns": [
        {"name": "batch_id", "type": "INTEGER", "primary_key": True},
        {"name": "operation_id", "type": "VARCHAR(100)"},
        {"name": "file_path", "type": "VARCHAR(500)", "nullable": True},
        {"name": "file_size_bytes", "type": "BIGINT", "nullable": True},
        {"name": "file_format", "type": "VARCHAR(50)", "nullable": True},
        {"name": "compression", "type": "VARCHAR(50)", "nullable": True},
        {"name": "row_count", "type": "INTEGER"},
        {"name": "checksum", "type": "VARCHAR(100)", "nullable": True},
        {"name": "created_at", "type": "TIMESTAMP"},
    ],
}

# All tables in write primitives schema
TABLES = {
    # Base TPC-H tables (read-only references)
    "region": REGION,
    "nation": NATION,
    "customer": CUSTOMER,
    "supplier": SUPPLIER,
    "part": PART,
    "partsupp": PARTSUPP,
    "orders": ORDERS,
    "lineitem": LINEITEM,
    # Category-based staging tables for write operations
    "insert_ops_lineitem": INSERT_OPS_LINEITEM,
    "insert_ops_orders": INSERT_OPS_ORDERS,
    "insert_ops_orders_summary": INSERT_OPS_ORDERS_SUMMARY,
    "insert_ops_lineitem_enriched": INSERT_OPS_LINEITEM_ENRICHED,
    "update_ops_orders": UPDATE_OPS_ORDERS,
    "delete_ops_orders": DELETE_OPS_ORDERS,
    "delete_ops_lineitem": DELETE_OPS_LINEITEM,
    "delete_ops_supplier": DELETE_OPS_SUPPLIER,
    "merge_ops_target": MERGE_OPS_TARGET,
    "merge_ops_source": MERGE_OPS_SOURCE,
    "merge_ops_lineitem_target": MERGE_OPS_LINEITEM_TARGET,
    "merge_ops_summary_target": MERGE_OPS_SUMMARY_TARGET,
    "bulk_load_ops_target": BULK_LOAD_OPS_TARGET,
    "ddl_truncate_target": DDL_TRUNCATE_TARGET,
    # Metadata tables
    "write_ops_log": WRITE_OPS_LOG,
    "batch_metadata": BATCH_METADATA,
}

# Tables that need to be created (excluding base TPC-H tables)
STAGING_TABLES = {
    "insert_ops_lineitem": INSERT_OPS_LINEITEM,
    "insert_ops_orders": INSERT_OPS_ORDERS,
    "insert_ops_orders_summary": INSERT_OPS_ORDERS_SUMMARY,
    "insert_ops_lineitem_enriched": INSERT_OPS_LINEITEM_ENRICHED,
    "update_ops_orders": UPDATE_OPS_ORDERS,
    "delete_ops_orders": DELETE_OPS_ORDERS,
    "delete_ops_lineitem": DELETE_OPS_LINEITEM,
    "delete_ops_supplier": DELETE_OPS_SUPPLIER,
    "merge_ops_target": MERGE_OPS_TARGET,
    "merge_ops_source": MERGE_OPS_SOURCE,
    "merge_ops_lineitem_target": MERGE_OPS_LINEITEM_TARGET,
    "merge_ops_summary_target": MERGE_OPS_SUMMARY_TARGET,
    "bulk_load_ops_target": BULK_LOAD_OPS_TARGET,
    "ddl_truncate_target": DDL_TRUNCATE_TARGET,
    "write_ops_log": WRITE_OPS_LOG,
    "batch_metadata": BATCH_METADATA,
}


def _supports_primary_keys(dialect: str) -> bool:
    """Return whether the target SQL dialect supports PRIMARY KEY in CREATE TABLE."""
    return dialect.lower() not in {"datafusion"}


def get_create_table_sql(table_name: str, dialect: str = "standard", if_not_exists: bool = False) -> str:
    """Generate CREATE TABLE SQL for a given table.

    Args:
        table_name: Name of the table to create
        dialect: SQL dialect to use (standard, postgres, mysql, etc.)
        if_not_exists: If True, use CREATE TABLE IF NOT EXISTS

    Returns:
        CREATE TABLE SQL statement

    Raises:
        ValueError: If table_name is not valid
    """
    if table_name not in STAGING_TABLES:
        raise ValueError(f"Unknown staging table: {table_name}. Use STAGING_TABLES only.")

    table = STAGING_TABLES[table_name]
    supports_primary_keys = _supports_primary_keys(dialect)
    columns: list[str] = []

    for col in table["columns"]:
        col_def = f"{col['name']} {col['type']}"
        if col.get("primary_key") and supports_primary_keys:
            col_def += " PRIMARY KEY"
        if not col.get("nullable", False) and not col.get("primary_key"):
            col_def += " NOT NULL"
        columns.append(col_def)

    # Handle composite primary keys
    if supports_primary_keys and "primary_key" in table and isinstance(table["primary_key"], list):
        pk_cols = ", ".join(table["primary_key"])
        columns.append(f"PRIMARY KEY ({pk_cols})")

    # Use IF NOT EXISTS if requested (supported by DuckDB, PostgreSQL, MySQL, SQLite)
    if_not_exists_clause = " IF NOT EXISTS" if if_not_exists else ""
    sql = f"CREATE TABLE{if_not_exists_clause} {table['name']} (\n"
    sql += ",\n".join(f"  {col}" for col in columns)
    sql += "\n);"

    return sql


def get_all_staging_tables_sql(dialect: str = "standard") -> str:
    """Generate CREATE TABLE SQL for all staging tables.

    Args:
        dialect: SQL dialect to use

    Returns:
        Complete SQL schema creation script for staging tables
    """
    # Order tables by dependencies (independent tables first)
    table_order = [
        "write_ops_log",
        "batch_metadata",
        "insert_ops_lineitem",
        "insert_ops_orders",
        "insert_ops_orders_summary",
        "insert_ops_lineitem_enriched",
        "update_ops_orders",
        "delete_ops_orders",
        "delete_ops_lineitem",
        "delete_ops_supplier",
        "merge_ops_target",
        "merge_ops_source",
        "merge_ops_lineitem_target",
        "merge_ops_summary_target",
        "bulk_load_ops_target",
        "ddl_truncate_target",
    ]

    sql_statements: list[str] = []
    for table_name in table_order:
        sql_statements.append(get_create_table_sql(table_name, dialect))

    return "\n\n".join(sql_statements)


def get_table_schema(table_name: str) -> dict[str, Any]:
    """Get schema definition for a table.

    Args:
        table_name: Name of the table

    Returns:
        Table schema dictionary

    Raises:
        ValueError: If table_name is not valid
    """
    if table_name not in TABLES:
        raise ValueError(f"Unknown table: {table_name}")
    return cast(dict[str, Any], TABLES[table_name])


__all__ = [
    "TABLES",
    "STAGING_TABLES",
    "INSERT_OPS_LINEITEM",
    "INSERT_OPS_ORDERS",
    "INSERT_OPS_ORDERS_SUMMARY",
    "INSERT_OPS_LINEITEM_ENRICHED",
    "UPDATE_OPS_ORDERS",
    "DELETE_OPS_ORDERS",
    "DELETE_OPS_LINEITEM",
    "DELETE_OPS_SUPPLIER",
    "MERGE_OPS_TARGET",
    "MERGE_OPS_SOURCE",
    "MERGE_OPS_LINEITEM_TARGET",
    "MERGE_OPS_SUMMARY_TARGET",
    "BULK_LOAD_OPS_TARGET",
    "DDL_TRUNCATE_TARGET",
    "WRITE_OPS_LOG",
    "BATCH_METADATA",
    "get_create_table_sql",
    "get_all_staging_tables_sql",
    "get_table_schema",
]
